import discretize
import numpy as np
import scipy.sparse as sp

from ...simulation import BaseSimulation
from ... import props
from ... import maps
from ...utils import (
    mkvc,
    validate_type,
    validate_ndarray_with_shape,
    validate_active_indices,
)

from .survey import SurveyVRM
from .receivers import Point, SquareLoop


############################################
# BASE VRM PROBLEM CLASS
############################################


class BaseVRMSimulation(BaseSimulation):
    """"""

    _AisSet = False

    def __init__(
        self,
        mesh,
        survey=None,
        refinement_factor=None,
        refinement_distance=None,
        indActive=None,
        **kwargs,
    ):
        super().__init__(mesh=mesh, survey=survey, **kwargs)

        if refinement_distance is None:
            if refinement_factor is None:
                refinement_factor = 3
            refinement_distance = (
                1.25
                * np.mean(
                    np.r_[np.min(mesh.h[0]), np.min(mesh.h[1]), np.min(mesh.h[2])]
                )
                * np.arange(1, refinement_factor + 1)
            )
        self.refinement_distance = refinement_distance
        if indActive is None:
            indActive = np.ones(self.mesh.n_cells, dtype=bool)
        self.indActive = indActive

    @BaseSimulation.mesh.setter
    def mesh(self, value):
        value = validate_type(
            "mesh", value, (discretize.TensorMesh, discretize.TreeMesh), cast=False
        )
        if value.dim != 3:
            raise ValueError(
                f"Mesh must be 3D tensor or 3D tree. Current mesh is {value.dim}"
            )
        self._mesh = value

    @property
    def survey(self):
        """The VRM survey.

        Returns
        -------
        SimPEG.electromagnetics.viscous_temanent_magnetization.survey.Survey
        """
        if self._survey is None:
            raise AttributeError("Simulation must have a survey.")
        return self._survey

    @survey.setter
    def survey(self, value):
        if value is not None:
            value = validate_type("survey", value, SurveyVRM, cast=False)
        self._survey = value

    @property
    def refinement_factor(self):
        """The number of refinement distances.

        Returns
        -------
        int
        """
        return len(self.refinement_distance)

    @property
    def refinement_distance(self):
        """Sensitivity refinement radii from sources.

        Returns
        -------
        numpy.ndarray
        """
        return self._refinement_distance

    @refinement_distance.setter
    def refinement_distance(self, value):
        self._refinement_distance = validate_ndarray_with_shape(
            "refinement_distance", value, shape=("*",)
        )

    @property
    def indActive(self):
        """Topography active cells.

        Returns
        -------
        (mesh.n_cells) numpy.ndarray of bool
        """
        return self._indActive

    @indActive.setter
    def indActive(self, value):
        self._indActive = validate_active_indices("indActive", value, self.mesh.n_cells)

    def _getH0matrix(self, xyz, pp):
        """
                Creates sparse matrix containing inducing field components
                for source pp

        ..        REQUIRED ARGUMENTS:
        ..
        ..        xyz: N X 3 array of locations to predict field
        ..
        ..        pp: Source index
        ..
        ..        OUTPUTS:
        ..
        ..        H0: A 3N X N sparse array containing Hx, Hy and Hz at all locations
        ..
        """

        srcObj = self.survey.source_list[pp]

        h0 = srcObj.getH0(xyz)

        hx0 = sp.diags(h0[:, 0], format="csr")
        hy0 = sp.diags(h0[:, 1], format="csr")
        hz0 = sp.diags(h0[:, 2], format="csr")

        h0 = sp.vstack([hx0, hy0, hz0])

        return h0

    def _getGeometryMatrix(self, xyzc, xyzh, pp):
        """
                Creates the dense geometry matrix which maps from the magnetized voxel
                cells to the receiver locations for source pp
        ..
        ..        REQUIRED ARGUMENTS:
        ..
        ..        xyzc: N by 3 numpy array containing cell center locations [xc,yc,zc]
        ..
        ..        xyzh: N by 3 numpy array containing cell dimensions [hx,hy,hz]
        ..
        ..        pp: Source index
        ..
        ..        OUTPUTS:
        ..
        ..        G: Linear geometry operator

        """

        srcObj = self.survey.source_list[pp]

        nC = np.shape(xyzc)[0]  # Number of cells
        nRx = srcObj.nRx  # Number of receiver in all receiver_list

        ax = np.reshape(xyzc[:, 0] - xyzh[:, 0] / 2, (1, nC))
        bx = np.reshape(xyzc[:, 0] + xyzh[:, 0] / 2, (1, nC))
        ay = np.reshape(xyzc[:, 1] - xyzh[:, 1] / 2, (1, nC))
        by = np.reshape(xyzc[:, 1] + xyzh[:, 1] / 2, (1, nC))
        az = np.reshape(xyzc[:, 2] - xyzh[:, 2] / 2, (1, nC))
        bz = np.reshape(xyzc[:, 2] + xyzh[:, 2] / 2, (1, nC))

        G = np.zeros((nRx, 3 * nC))
        c = -(1 / (4 * np.pi))
        tol = 1e-10  # Tolerance constant for numerical stability
        tol2 = 1000.0  # Tolerance constant for numerical stability

        COUNT = 0

        for qq in range(0, len(srcObj.receiver_list)):
            rxObj = srcObj.receiver_list[qq]
            dComp = rxObj.orientation
            locs = rxObj.locations
            nLoc = np.shape(locs)[0]

            if isinstance(rxObj, Point) and not isinstance(rxObj, SquareLoop):
                if dComp.lower() == "x":
                    for rr in range(0, nLoc):
                        u1 = locs[rr, 0] - ax
                        u1[np.abs(u1) < tol] = np.min(xyzh[:, 0]) / tol2
                        u2 = locs[rr, 0] - bx
                        u2[np.abs(u2) < tol] = -np.min(xyzh[:, 0]) / tol2
                        v1 = locs[rr, 1] - ay
                        v1[np.abs(v1) < tol] = np.min(xyzh[:, 1]) / tol2
                        v2 = locs[rr, 1] - by
                        v2[np.abs(v2) < tol] = -np.min(xyzh[:, 1]) / tol2
                        w1 = locs[rr, 2] - az
                        w1[np.abs(w1) < tol] = np.min(xyzh[:, 2]) / tol2
                        w2 = locs[rr, 2] - bz
                        w2[np.abs(w2) < tol] = -np.min(xyzh[:, 2]) / tol2

                        d111 = np.sqrt(u1**2 + v1**2 + w1**2)
                        d211 = np.sqrt(u2**2 + v1**2 + w1**2)
                        d221 = np.sqrt(u2**2 + v2**2 + w1**2)
                        d121 = np.sqrt(u1**2 + v2**2 + w1**2)
                        d122 = np.sqrt(u1**2 + v2**2 + w2**2)
                        d112 = np.sqrt(u1**2 + v1**2 + w2**2)
                        d212 = np.sqrt(u2**2 + v1**2 + w2**2)
                        d222 = np.sqrt(u2**2 + v2**2 + w2**2)

                        Gxx = (
                            np.arctan((v1 * w1) / (u1 * d111 + tol))
                            - np.arctan((v1 * w1) / (u2 * d211 + tol))
                            + np.arctan((v2 * w1) / (u2 * d221 + tol))
                            - np.arctan((v2 * w1) / (u1 * d121 + tol))
                            + np.arctan((v2 * w2) / (u1 * d122 + tol))
                            - np.arctan((v1 * w2) / (u1 * d112 + tol))
                            + np.arctan((v1 * w2) / (u2 * d212 + tol))
                            - np.arctan((v2 * w2) / (u2 * d222 + tol))
                        )

                        Gyx = (
                            np.log(d111 - w1)
                            - np.log(d211 - w1)
                            + np.log(d221 - w1)
                            - np.log(d121 - w1)
                            + np.log(d122 - w2)
                            - np.log(d112 - w2)
                            + np.log(d212 - w2)
                            - np.log(d222 - w2)
                        )

                        Gzx = (
                            np.log(d111 - v1)
                            - np.log(d211 - v1)
                            + np.log(d221 - v2)
                            - np.log(d121 - v2)
                            + np.log(d122 - v2)
                            - np.log(d112 - v1)
                            + np.log(d212 - v1)
                            - np.log(d222 - v2)
                        )

                        G[COUNT, :] = c * np.c_[Gxx, Gyx, Gzx]
                        COUNT = COUNT + 1

                elif dComp.lower() == "y":
                    for rr in range(0, nLoc):
                        u1 = locs[rr, 0] - ax
                        u1[np.abs(u1) < tol] = np.min(xyzh[:, 0]) / tol2
                        u2 = locs[rr, 0] - bx
                        u2[np.abs(u2) < tol] = -np.min(xyzh[:, 0]) / tol2
                        v1 = locs[rr, 1] - ay
                        v1[np.abs(v1) < tol] = np.min(xyzh[:, 1]) / tol2
                        v2 = locs[rr, 1] - by
                        v2[np.abs(v2) < tol] = -np.min(xyzh[:, 1]) / tol2
                        w1 = locs[rr, 2] - az
                        w1[np.abs(w1) < tol] = np.min(xyzh[:, 2]) / tol2
                        w2 = locs[rr, 2] - bz
                        w2[np.abs(w2) < tol] = -np.min(xyzh[:, 2]) / tol2

                        d111 = np.sqrt(u1**2 + v1**2 + w1**2)
                        d211 = np.sqrt(u2**2 + v1**2 + w1**2)
                        d221 = np.sqrt(u2**2 + v2**2 + w1**2)
                        d121 = np.sqrt(u1**2 + v2**2 + w1**2)
                        d122 = np.sqrt(u1**2 + v2**2 + w2**2)
                        d112 = np.sqrt(u1**2 + v1**2 + w2**2)
                        d212 = np.sqrt(u2**2 + v1**2 + w2**2)
                        d222 = np.sqrt(u2**2 + v2**2 + w2**2)

                        Gxy = (
                            np.log(d111 - w1)
                            - np.log(d211 - w1)
                            + np.log(d221 - w1)
                            - np.log(d121 - w1)
                            + np.log(d122 - w2)
                            - np.log(d112 - w2)
                            + np.log(d212 - w2)
                            - np.log(d222 - w2)
                        )

                        Gyy = (
                            np.arctan((u1 * w1) / (v1 * d111 + tol))
                            - np.arctan((u2 * w1) / (v1 * d211 + tol))
                            + np.arctan((u2 * w1) / (v2 * d221 + tol))
                            - np.arctan((u1 * w1) / (v2 * d121 + tol))
                            + np.arctan((u1 * w2) / (v2 * d122 + tol))
                            - np.arctan((u1 * w2) / (v1 * d112 + tol))
                            + np.arctan((u2 * w2) / (v1 * d212 + tol))
                            - np.arctan((u2 * w2) / (v2 * d222 + tol))
                        )

                        Gzy = (
                            np.log(d111 - u1)
                            - np.log(d211 - u2)
                            + np.log(d221 - u2)
                            - np.log(d121 - u1)
                            + np.log(d122 - u1)
                            - np.log(d112 - u1)
                            + np.log(d212 - u2)
                            - np.log(d222 - u2)
                        )

                        G[COUNT, :] = c * np.c_[Gxy, Gyy, Gzy]
                        COUNT = COUNT + 1

                elif dComp.lower() == "z":
                    for rr in range(0, nLoc):
                        u1 = locs[rr, 0] - ax
                        u1[np.abs(u1) < tol] = np.min(xyzh[:, 0]) / tol2
                        u2 = locs[rr, 0] - bx
                        u2[np.abs(u2) < tol] = -np.min(xyzh[:, 0]) / tol2
                        v1 = locs[rr, 1] - ay
                        v1[np.abs(v1) < tol] = np.min(xyzh[:, 1]) / tol2
                        v2 = locs[rr, 1] - by
                        v2[np.abs(v2) < tol] = -np.min(xyzh[:, 1]) / tol2
                        w1 = locs[rr, 2] - az
                        w1[np.abs(w1) < tol] = np.min(xyzh[:, 2]) / tol2
                        w2 = locs[rr, 2] - bz
                        w2[np.abs(w2) < tol] = -np.min(xyzh[:, 2]) / tol2

                        d111 = np.sqrt(u1**2 + v1**2 + w1**2)
                        d211 = np.sqrt(u2**2 + v1**2 + w1**2)
                        d221 = np.sqrt(u2**2 + v2**2 + w1**2)
                        d121 = np.sqrt(u1**2 + v2**2 + w1**2)
                        d122 = np.sqrt(u1**2 + v2**2 + w2**2)
                        d112 = np.sqrt(u1**2 + v1**2 + w2**2)
                        d212 = np.sqrt(u2**2 + v1**2 + w2**2)
                        d222 = np.sqrt(u2**2 + v2**2 + w2**2)

                        Gxz = (
                            np.log(d111 - v1)
                            - np.log(d211 - v1)
                            + np.log(d221 - v2)
                            - np.log(d121 - v2)
                            + np.log(d122 - v2)
                            - np.log(d112 - v1)
                            + np.log(d212 - v1)
                            - np.log(d222 - v2)
                        )

                        Gyz = (
                            np.log(d111 - u1)
                            - np.log(d211 - u2)
                            + np.log(d221 - u2)
                            - np.log(d121 - u1)
                            + np.log(d122 - u1)
                            - np.log(d112 - u1)
                            + np.log(d212 - u2)
                            - np.log(d222 - u2)
                        )

                        Gzz = (
                            -np.arctan((v1 * w1) / (u1 * d111 + tol))
                            + np.arctan((v1 * w1) / (u2 * d211 + tol))
                            - np.arctan((v2 * w1) / (u2 * d221 + tol))
                            + np.arctan((v2 * w1) / (u1 * d121 + tol))
                            - np.arctan((v2 * w2) / (u1 * d122 + tol))
                            + np.arctan((v1 * w2) / (u1 * d112 + tol))
                            - np.arctan((v1 * w2) / (u2 * d212 + tol))
                            + np.arctan((v2 * w2) / (u2 * d222 + tol))
                        )

                        Gzz = (
                            Gzz
                            - np.arctan((u1 * w1) / (v1 * d111 + tol))
                            + np.arctan((u2 * w1) / (v1 * d211 + tol))
                            - np.arctan((u2 * w1) / (v2 * d221 + tol))
                            + np.arctan((u1 * w1) / (v2 * d121 + tol))
                            - np.arctan((u1 * w2) / (v2 * d122 + tol))
                            + np.arctan((u1 * w2) / (v1 * d112 + tol))
                            - np.arctan((u2 * w2) / (v1 * d212 + tol))
                            + np.arctan((u2 * w2) / (v2 * d222 + tol))
                        )

                        G[COUNT, :] = c * np.c_[Gxz, Gyz, Gzz]
                        COUNT = COUNT + 1

            elif isinstance(rxObj, SquareLoop):
                # Gaussian quadrature weights
                wt = [
                    np.r_[2.0],
                    np.r_[1.0, 1.0],
                    np.r_[0.555556, 0.888889, 0.555556],
                    np.r_[0.347855, 0.652145, 0.652145, 0.347855],
                    np.r_[0.236927, 0.478629, 0.568889, 0.478629, 0.236927],
                    np.r_[0.171324, 0.467914, 0.360762, 0.360762, 0.467914, 0.171324],
                    np.r_[
                        0.129485,
                        0.279705,
                        0.381830,
                        0.417959,
                        0.381830,
                        0.279705,
                        0.129485,
                    ],
                ]
                wt = wt[rxObj.quadrature_order - 1]
                nw = len(wt)
                wt = (
                    rxObj.n_turns
                    * (rxObj.width / 2) ** 2
                    * np.reshape(np.outer(wt, wt), (1, nw**2))
                )

                # Gaussian quadrature locations on [-1,1]
                ds = [
                    np.r_[0.0],
                    np.r_[-0.57735, 0.57735],
                    np.r_[-0.774597, 0.0, 0.774597],
                    np.r_[-0.861136, -0.339981, 0.339981, 0.861136],
                    np.r_[-0.906180, -0.538469, 0, 0.538469, 0.906180],
                    np.r_[
                        -0.932470, -0.238619, -0.661209, 0.661209, 0.238619, 0.932470
                    ],
                    np.r_[
                        -0.949108,
                        -0.741531,
                        -0.405845,
                        0.0,
                        0.405845,
                        0.741531,
                        0.949108,
                    ],
                ]

                s1 = (
                    0.5
                    * rxObj.width
                    * np.reshape(
                        np.kron(ds[rxObj.quadrature_order - 1], np.ones(nw)),
                        (nw**2, 1),
                    )
                )
                s2 = (
                    0.5
                    * rxObj.width
                    * np.reshape(
                        np.kron(np.ones(nw), ds[rxObj.quadrature_order - 1]),
                        (nw**2, 1),
                    )
                )

                if dComp.lower() == "x":
                    for rr in range(0, nLoc):
                        u1 = np.kron(np.ones((nw**2, 1)), locs[rr, 0] - ax)
                        u1[np.abs(u1) < tol] = np.min(xyzh[:, 0]) / tol2
                        u2 = np.kron(np.ones((nw**2, 1)), locs[rr, 0] - bx)
                        u2[np.abs(u2) < tol] = -np.min(xyzh[:, 0]) / tol2

                        v1 = np.kron(np.ones((nw**2, 1)), locs[rr, 1] - ay) + np.kron(
                            s1, np.ones((1, nC))
                        )
                        v1[np.abs(v1) < tol] = np.min(xyzh[:, 1]) / tol2
                        v2 = np.kron(np.ones((nw**2, 1)), locs[rr, 1] - by) + np.kron(
                            s1, np.ones((1, nC))
                        )
                        v2[np.abs(v2) < tol] = -np.min(xyzh[:, 1]) / tol2

                        w1 = np.kron(np.ones((nw**2, 1)), locs[rr, 2] - az) + np.kron(
                            s2, np.ones((1, nC))
                        )
                        w1[np.abs(w1) < tol] = np.min(xyzh[:, 2]) / tol2
                        w2 = np.kron(np.ones((nw**2, 1)), locs[rr, 2] - bz) + np.kron(
                            s2, np.ones((1, nC))
                        )
                        w2[np.abs(w2) < tol] = -np.min(xyzh[:, 2]) / tol2

                        d111 = np.sqrt(u1**2 + v1**2 + w1**2)
                        d211 = np.sqrt(u2**2 + v1**2 + w1**2)
                        d221 = np.sqrt(u2**2 + v2**2 + w1**2)
                        d121 = np.sqrt(u1**2 + v2**2 + w1**2)
                        d122 = np.sqrt(u1**2 + v2**2 + w2**2)
                        d112 = np.sqrt(u1**2 + v1**2 + w2**2)
                        d212 = np.sqrt(u2**2 + v1**2 + w2**2)
                        d222 = np.sqrt(u2**2 + v2**2 + w2**2)

                        Gxx = (
                            np.arctan((v1 * w1) / (u1 * d111 + tol))
                            - np.arctan((v1 * w1) / (u2 * d211 + tol))
                            + np.arctan((v2 * w1) / (u2 * d221 + tol))
                            - np.arctan((v2 * w1) / (u1 * d121 + tol))
                            + np.arctan((v2 * w2) / (u1 * d122 + tol))
                            - np.arctan((v1 * w2) / (u1 * d112 + tol))
                            + np.arctan((v1 * w2) / (u2 * d212 + tol))
                            - np.arctan((v2 * w2) / (u2 * d222 + tol))
                        )

                        Gxx = np.dot(wt, Gxx)

                        Gyx = (
                            np.log(d111 - w1)
                            - np.log(d211 - w1)
                            + np.log(d221 - w1)
                            - np.log(d121 - w1)
                            + np.log(d122 - w2)
                            - np.log(d112 - w2)
                            + np.log(d212 - w2)
                            - np.log(d222 - w2)
                        )

                        Gyx = np.dot(wt, Gyx)

                        Gzx = (
                            np.log(d111 - v1)
                            - np.log(d211 - v1)
                            + np.log(d221 - v2)
                            - np.log(d121 - v2)
                            + np.log(d122 - v2)
                            - np.log(d112 - v1)
                            + np.log(d212 - v1)
                            - np.log(d222 - v2)
                        )

                        Gzx = np.dot(wt, Gzx)

                        G[COUNT, :] = c * np.c_[Gxx, Gyx, Gzx]
                        COUNT = COUNT + 1

                elif dComp.lower() == "y":
                    for rr in range(0, nLoc):
                        u1 = np.kron(np.ones((nw**2, 1)), locs[rr, 0] - ax) + np.kron(
                            s1, np.ones((1, nC))
                        )
                        u1[np.abs(u1) < tol] = np.min(xyzh[:, 0]) / tol2
                        u2 = np.kron(np.ones((nw**2, 1)), locs[rr, 0] - bx) + np.kron(
                            s1, np.ones((1, nC))
                        )
                        u2[np.abs(u2) < tol] = -np.min(xyzh[:, 0]) / tol2

                        v1 = np.kron(np.ones((nw**2, 1)), locs[rr, 1] - ay)
                        v1[np.abs(v1) < tol] = np.min(xyzh[:, 1]) / tol2
                        v2 = np.kron(np.ones((nw**2, 1)), locs[rr, 1] - by)
                        v2[np.abs(v2) < tol] = -np.min(xyzh[:, 1]) / tol2

                        w1 = np.kron(np.ones((nw**2, 1)), locs[rr, 2] - az) + np.kron(
                            s2, np.ones((1, nC))
                        )
                        w1[np.abs(w1) < tol] = np.min(xyzh[:, 2]) / tol2
                        w2 = np.kron(np.ones((nw**2, 1)), locs[rr, 2] - bz) + np.kron(
                            s2, np.ones((1, nC))
                        )
                        w2[np.abs(w2) < tol] = -np.min(xyzh[:, 2]) / tol2

                        d111 = np.sqrt(u1**2 + v1**2 + w1**2)
                        d211 = np.sqrt(u2**2 + v1**2 + w1**2)
                        d221 = np.sqrt(u2**2 + v2**2 + w1**2)
                        d121 = np.sqrt(u1**2 + v2**2 + w1**2)
                        d122 = np.sqrt(u1**2 + v2**2 + w2**2)
                        d112 = np.sqrt(u1**2 + v1**2 + w2**2)
                        d212 = np.sqrt(u2**2 + v1**2 + w2**2)
                        d222 = np.sqrt(u2**2 + v2**2 + w2**2)

                        Gxy = (
                            np.log(d111 - w1)
                            - np.log(d211 - w1)
                            + np.log(d221 - w1)
                            - np.log(d121 - w1)
                            + np.log(d122 - w2)
                            - np.log(d112 - w2)
                            + np.log(d212 - w2)
                            - np.log(d222 - w2)
                        )

                        Gxy = np.dot(wt, Gxy)

                        Gyy = (
                            np.arctan((u1 * w1) / (v1 * d111 + tol))
                            - np.arctan((u2 * w1) / (v1 * d211 + tol))
                            + np.arctan((u2 * w1) / (v2 * d221 + tol))
                            - np.arctan((u1 * w1) / (v2 * d121 + tol))
                            + np.arctan((u1 * w2) / (v2 * d122 + tol))
                            - np.arctan((u1 * w2) / (v1 * d112 + tol))
                            + np.arctan((u2 * w2) / (v1 * d212 + tol))
                            - np.arctan((u2 * w2) / (v2 * d222 + tol))
                        )

                        Gyy = np.dot(wt, Gyy)

                        Gzy = (
                            np.log(d111 - u1)
                            - np.log(d211 - u2)
                            + np.log(d221 - u2)
                            - np.log(d121 - u1)
                            + np.log(d122 - u1)
                            - np.log(d112 - u1)
                            + np.log(d212 - u2)
                            - np.log(d222 - u2)
                        )

                        Gzy = np.dot(wt, Gzy)

                        G[COUNT, :] = c * np.c_[Gxy, Gyy, Gzy]
                        COUNT = COUNT + 1

                elif dComp.lower() == "z":
                    for rr in range(0, nLoc):
                        u1 = np.kron(np.ones((nw**2, 1)), locs[rr, 0] - ax) + np.kron(
                            s1, np.ones((1, nC))
                        )
                        u1[np.abs(u1) < tol] = np.min(xyzh[:, 0]) / tol2
                        u2 = np.kron(np.ones((nw**2, 1)), locs[rr, 0] - bx) + np.kron(
                            s1, np.ones((1, nC))
                        )

                        u2[np.abs(u2) < tol] = -np.min(xyzh[:, 0]) / tol2
                        v1 = np.kron(np.ones((nw**2, 1)), locs[rr, 1] - ay) + np.kron(
                            s2, np.ones((1, nC))
                        )
                        v1[np.abs(v1) < tol] = np.min(xyzh[:, 1]) / tol2
                        v2 = np.kron(np.ones((nw**2, 1)), locs[rr, 1] - by) + np.kron(
                            s2, np.ones((1, nC))
                        )
                        v2[np.abs(v2) < tol] = -np.min(xyzh[:, 1]) / tol2

                        w1 = np.kron(np.ones((nw**2, 1)), locs[rr, 2] - az)
                        w1[np.abs(w1) < tol] = np.min(xyzh[:, 2]) / tol2
                        w2 = np.kron(np.ones((nw**2, 1)), locs[rr, 2] - bz)
                        w2[np.abs(w2) < tol] = -np.min(xyzh[:, 2]) / tol2

                        d111 = np.sqrt(u1**2 + v1**2 + w1**2)
                        d211 = np.sqrt(u2**2 + v1**2 + w1**2)
                        d221 = np.sqrt(u2**2 + v2**2 + w1**2)
                        d121 = np.sqrt(u1**2 + v2**2 + w1**2)
                        d122 = np.sqrt(u1**2 + v2**2 + w2**2)
                        d112 = np.sqrt(u1**2 + v1**2 + w2**2)
                        d212 = np.sqrt(u2**2 + v1**2 + w2**2)
                        d222 = np.sqrt(u2**2 + v2**2 + w2**2)

                        Gxz = (
                            np.log(d111 - v1)
                            - np.log(d211 - v1)
                            + np.log(d221 - v2)
                            - np.log(d121 - v2)
                            + np.log(d122 - v2)
                            - np.log(d112 - v1)
                            + np.log(d212 - v1)
                            - np.log(d222 - v2)
                        )

                        Gxz = np.dot(wt, Gxz)

                        Gyz = (
                            np.log(d111 - u1)
                            - np.log(d211 - u2)
                            + np.log(d221 - u2)
                            - np.log(d121 - u1)
                            + np.log(d122 - u1)
                            - np.log(d112 - u1)
                            + np.log(d212 - u2)
                            - np.log(d222 - u2)
                        )

                        Gyz = np.dot(wt, Gyz)

                        Gzz = (
                            -np.arctan((v1 * w1) / (u1 * d111 + tol))
                            + np.arctan((v1 * w1) / (u2 * d211 + tol))
                            - np.arctan((v2 * w1) / (u2 * d221 + tol))
                            + np.arctan((v2 * w1) / (u1 * d121 + tol))
                            - np.arctan((v2 * w2) / (u1 * d122 + tol))
                            + np.arctan((v1 * w2) / (u1 * d112 + tol))
                            - np.arctan((v1 * w2) / (u2 * d212 + tol))
                            + np.arctan((v2 * w2) / (u2 * d222 + tol))
                        )

                        Gzz = (
                            Gzz
                            - np.arctan((u1 * w1) / (v1 * d111 + tol))
                            + np.arctan((u2 * w1) / (v1 * d211 + tol))
                            - np.arctan((u2 * w1) / (v2 * d221 + tol))
                            + np.arctan((u1 * w1) / (v2 * d121 + tol))
                            - np.arctan((u1 * w2) / (v2 * d122 + tol))
                            + np.arctan((u1 * w2) / (v1 * d112 + tol))
                            - np.arctan((u2 * w2) / (v1 * d212 + tol))
                            + np.arctan((u2 * w2) / (v2 * d222 + tol))
                        )

                        Gzz = np.dot(wt, Gzz)

                        G[COUNT, :] = c * np.c_[Gxz, Gyz, Gzz]
                        COUNT = COUNT + 1

        return G

    def _getAMatricies(self):
        """Returns the full geometric operator"""

        indActive = self.indActive

        # GET CELL INFORMATION FOR FORWARD MODELING
        meshObj = self.mesh
        xyzc = meshObj.gridCC[indActive, :]
        xyzh = meshObj.h_gridded[indActive, :]

        # GET LIST OF A MATRICIES
        A = []
        for pp in range(0, self.survey.nSrc):
            # Create initial A matrix
            G = self._getGeometryMatrix(xyzc, xyzh, pp)
            H0 = self._getH0matrix(xyzc, pp)
            A.append(G * H0)

            # Refine A matrix
            refinement_factor = self.refinement_factor
            refinement_distance = self.refinement_distance

            if refinement_factor > 0:
                srcObj = self.survey.source_list[pp]
                refFlag = srcObj._getRefineFlags(
                    xyzc, refinement_factor, refinement_distance
                )

                for qq in range(1, refinement_factor + 1):
                    if len(refFlag[refFlag == qq]) != 0:
                        A[pp][:, refFlag == qq] = self._getSubsetAcolumns(
                            xyzc, xyzh, pp, qq, refFlag
                        )

        return A

    def _getSubsetAcolumns(self, xyzc, xyzh, pp, qq, refFlag):
        """
                This method returns the refined sensitivities for columns that will be
                replaced in the A matrix for source pp and refinement factor qq.
        ..
        ..        INPUTS:
        ..
        ..        xyzc -- Cell centers of topo mesh cells N X 3 array
        ..
        ..        xyzh -- Cell widths of topo mesh cells N X 3 array
        ..
        ..        pp -- Source ID
        ..
        ..        qq -- Mesh refinement factor
        ..
        ..        refFlag -- refinement factors for all topo mesh cells
        ..
        ..        OUTPUTS:
        ..
        ..        Acols -- Columns containing replacement sensitivities

        """

        # GET SUBMESH GRID
        n = 2**qq
        [nx, ny, nz] = np.meshgrid(
            np.linspace(1, n, n) - 0.5,
            np.linspace(1, n, n) - 0.5,
            np.linspace(1, n, n) - 0.5,
        )
        nxyz_sub = np.c_[mkvc(nx), mkvc(ny), mkvc(nz)]

        xyzh_sub = xyzh[refFlag == qq, :]  # Get widths of cells to be refined
        xyzc_sub = (
            xyzc[refFlag == qq, :] - xyzh[refFlag == qq, :] / 2
        )  # Get bottom southwest corners of cells to be refined
        m = np.shape(xyzc_sub)[0]
        xyzc_sub = np.kron(
            xyzc_sub, np.ones((n**3, 1))
        )  # Kron for n**3 refined cells
        xyzh_sub = np.kron(
            xyzh_sub / n, np.ones((n**3, 1))
        )  # Kron for n**3 refined cells with widths h/n
        nxyz_sub = np.kron(np.ones((m, 1)), nxyz_sub)  # Kron for n**3 refined cells
        xyzc_sub = xyzc_sub + xyzh_sub * nxyz_sub

        # GET SUBMESH A MATRIX AND COLLAPSE TO COLUMNS
        G = self._getGeometryMatrix(xyzc_sub, xyzh_sub, pp)
        H0 = self._getH0matrix(xyzc_sub, pp)
        Acols = (G * H0) * sp.kron(sp.diags(np.ones(m)), np.ones((n**3, 1)))

        return Acols

    def dpred(self, m=None, f=None):
        """"""
        if f is None:
            f = self.fields(m)

        return f[self.survey.t_active]


#############################################################################
# VRM CHARACTERISTIC DECAY FORMULATION (SINGLE MODEL PARAMETER AND INVERSION)
#############################################################################


class Simulation3DLinear(BaseVRMSimulation):
    """"""

    _A = None
    _T = None
    _TisSet = False
    _xiMap = None

    xi, xiMap, xiDeriv = props.Invertible(
        "Amalgamated Viscous Remanent Magnetization Parameter xi = dchi/ln(tau2/tau1)"
    )

    def __init__(self, mesh, xi=None, xiMap=None, **kwargs):
        super().__init__(mesh, **kwargs)
        self.xi = xi
        self.xiMap = xiMap

        nAct = list(self.indActive).count(True)
        if self.xiMap is None:
            self.xiMap = maps.IdentityMap(nP=nAct)

    @property
    def A(self):
        """
        The geometric sensitivity matrix for the linear VRM problem. Accessing
        this property requires that the problem be paired with a survey object.

        """

        if self._AisSet is False:
            if self.survey is None:
                AssertionError("A survey must be set to generate A matrix")

            # Remove any previously stored A matrix
            if self._A is not None:
                self._A = None

            print("CREATING A MATRIX")

            # COLLAPSE ALL A MATRICIES INTO SINGLE OPERATOR
            self._A = np.vstack(self._getAMatricies())
            self._AisSet = True

            return self._A

        elif self._AisSet is True:
            return self._A

    @property
    def T(self):
        """
        The characteristic decay matrix for the VRM problem. Accessing this
        property requires that the problem be paired with a survey object.

        """

        if self._TisSet is False:
            if self.survey is None:
                AssertionError("A survey must be set to generate A matrix")

            # Remove any previously stored T matrix
            if self._T is not None:
                self._T = None

            print("CREATING T MATRIX")

            source_list = self.survey.source_list
            nSrc = len(source_list)
            T = []

            for pp in range(0, nSrc):
                receiver_list = source_list[pp].receiver_list
                nRx = len(receiver_list)
                waveObj = source_list[pp].waveform

                for qq in range(0, nRx):
                    times = receiver_list[qq].times
                    nLoc = np.shape(receiver_list[qq].locations)[0]

                    I = sp.diags(np.ones(nLoc))
                    eta = waveObj.getCharDecay(receiver_list[qq].field_type, times)
                    eta = np.atleast_2d(eta).T

                    T.append(sp.kron(I, eta))

            self._T = sp.block_diag(T)
            self._TisSet = True

            return self._T

        elif self._TisSet is True:
            return self._T

    def fields(self, m):
        """Computes the fields d = T*A*m"""

        if self.survey is None:
            AssertionError("A survey must be set to generate A matrix")

        self.model = m  # Initiates/updates model and initiates mapping

        # Project to active mesh cells
        m = self.xiMap * m

        # Must return as a numpy array
        return mkvc(sp.coo_matrix.dot(self.T, np.dot(self.A, m)))

    def Jvec(self, m, v, f=None):
        """Compute Pd*T*A*dxidm*v"""

        if self.survey is None:
            AssertionError("A survey must be set to generate A matrix")

        # Jacobian of xi wrt model
        dxidm = self.xiMap.deriv(m)

        # dxidm*v
        v = dxidm * v

        # Dot product with A
        v = self.A @ v

        # Get active time rows of T
        T = self.T.tocsr()[self.survey.t_active, :]

        # Must return an array
        return mkvc(sp.csr_matrix.dot(T, v))

    def Jtvec(self, m, v, f=None):
        """Compute (Pd*T*A*dxidm)^T * v"""

        if self.survey is None:
            AssertionError("A survey must be set to generate A matrix")

        # Define v as a column vector
        v = np.atleast_2d(v).T

        # Get T'*Pd'*v
        T = self.T.tocsr()[self.survey.t_active, :]
        v = sp.csc_matrix.dot(T.transpose(), v)

        # Multiply by A'
        v = (np.dot(v.T, self.A)).T

        # Jacobian of xi wrt model
        dxidm = self.xiMap.deriv(m)

        # Must return an array
        return mkvc(dxidm.T * v)


class Simulation3DLogUniform(BaseVRMSimulation):
    """"""

    _A = None
    _T = None
    _TisSet = False
    # _xiMap = None

    chi0 = props.PhysicalProperty("DC susceptibility")
    dchi = props.PhysicalProperty("Frequency dependence")
    tau1 = props.PhysicalProperty("Low bound time-relaxation constant")
    tau2 = props.PhysicalProperty("Upper bound time-relaxation constant")

    def __init__(
        self, mesh, survey=None, chi0=None, dchi=None, tau1=None, tau2=None, **kwargs
    ):
        super(Simulation3DLogUniform, self).__init__(mesh=mesh, survey=survey, **kwargs)
        self.chi0 = chi0
        self.dchi = dchi
        self.tau1 = tau1
        self.tau2 = tau2

    @property
    def A(self):
        """
        The geometric sensitivity matrix for the linear VRM problem. Accessing
        this property requires that the problem be paired with a survey object.

        """

        if self._AisSet is False:
            if self.survey is None:
                AssertionError("A survey must be set to generate A matrix")

            # Remove any previously stored A matrix
            if self._A is not None:
                self._A = None

            print("CREATING A MATRIX")

            # COLLAPSE ALL A MATRICIES INTO SINGLE OPERATOR
            self._A = self._getAMatricies()
            self._AisSet = True

            return self._A

        elif self._AisSet is True:
            return self._A

    def fields(self, m=None):
        """Computes the fields at every time d(t) = G*M(t)"""

        if self.survey is None:
            AssertionError("A survey must be set to generate A matrix")

        # Fields from each source
        source_list = self.survey.source_list
        nSrc = len(source_list)
        f = []

        for pp in range(0, nSrc):
            receiver_list = source_list[pp].receiver_list
            nRx = len(receiver_list)
            waveObj = source_list[pp].waveform

            for qq in range(0, nRx):
                times = receiver_list[qq].times
                eta = waveObj.getLogUniformDecay(
                    receiver_list[qq].field_type,
                    times,
                    self.chi0,
                    self.dchi,
                    self.tau1,
                    self.tau2,
                )

                f.append(mkvc(self.A[qq] @ eta))

        return np.array(np.hstack(f))



import numpy as np
import scipy.sparse as sp
from discretize.utils import Zero
from ..simulation import BaseSimulation
from .. import props
from scipy.constants import mu_0


def __inner_mat_mul_op(M, u, v=None, adjoint=False):
    u = np.squeeze(u)
    if v is not None:
        if u.ndim > 1:
            if u.shape[1] > 1 and not isinstance(u, (sp.csr_matrix, sp.csc_matrix, sp.coo_matrix)):
                # u has multiple fields
                if v.ndim == 1:
                    v = sp.diags(v)
            else:
                u = u[:, 0]
        if u.ndim == 1:
            if v.ndim > 1:
                if v.shape[1] == 1:
                    v = v[:, 0]
                else:
                    u = sp.diags(u)
        if v.ndim > 2:
            u = u[:, :, None]
        if adjoint:
            if u.ndim > 1 and u.shape[1] > 1 and not isinstance(u, sp.dia_matrix):
                return M.T * (
                        u[:, None, :] *
                        v.reshape((u.shape[0], -1, u.shape[1]))
                ).sum(axis=2)
            return M.T * (u * v)
        if u.ndim > 1 and u.shape[1] > 1:
            return np.squeeze(u[:, None, :] * (M * v)[:, :, None])
        return u * (M * v)
    else:
        if u.ndim > 1:
            UM = sp.vstack([sp.diags(u[:, i]) @ M for i in range(u.shape[1])])
        else:
            U = sp.diags(u, format="csr")
            UM = U @ M
        if adjoint:
            return UM.T
        return UM


def with_property_mass_matrices(property_name):
    """
    This decorator will automatically populate all of the property mass matrices.

    Given the property "prop", this will add properties and functions to the class
    representing all of the possible mass matrix operations on the mesh.

    For a given property, "prop", they will be named:

    * MccProp
    * MccPropDeriv
    * MccPropI
    * MccPropIDeriv

    and so on for each "Mcc", "Mn", "Mf", and "Me".
    """

    def decorator(cls):
        arg = property_name.lower()
        arg = arg[0].upper() + arg[1:]

        @property
        def Mcc_prop(self):
            """
            Cell center property inner product matrix.
            """
            stash_name = f"_Mcc_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                M_prop = sp.diags(self.mesh.cell_volumes * prop, format="csr")
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Mcc{arg}", Mcc_prop)

        @property
        def Mn_prop(self):
            """
            Node property inner product matrix.
            """
            stash_name = f"_Mn_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                vol = self.mesh.cell_volumes
                M_prop = sp.diags(self.mesh.aveN2CC.T * (vol * prop), format="csr")
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Mn{arg}", Mn_prop)

        @property
        def Mf_prop(self):
            """
            Face property inner product matrix.
            """
            stash_name = f"_Mf_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                M_prop = self.mesh.get_face_inner_product(model=prop)
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Mf{arg}", Mf_prop)

        @property
        def Me_prop(self):
            """
            Edge property inner product matrix.
            """
            stash_name = f"_Me_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                M_prop = self.mesh.get_edge_inner_product(model=prop)
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Me{arg}", Me_prop)

        @property
        def MccI_prop(self):
            """
            Cell center property inner product inverse matrix.
            """
            stash_name = f"_MccI_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                M_prop = sp.diags(1.0 / (self.mesh.cell_volumes * prop), format="csr")
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Mcc{arg}I", MccI_prop)

        @property
        def MnI_prop(self):
            """
            Node property inner product inverse matrix.
            """
            stash_name = f"_MnI_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                vol = self.mesh.cell_volumes
                M_prop = sp.diags(
                    1.0 / (self.mesh.aveN2CC.T * (vol * prop)), format="csr"
                )
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Mn{arg}I", MnI_prop)

        @property
        def MfI_prop(self):
            """
            Face property inner product inverse matrix.
            """
            stash_name = f"_MfI_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                M_prop = self.mesh.get_face_inner_product(
                    model=prop, invert_matrix=True
                )
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Mf{arg}I", MfI_prop)

        @property
        def MeI_prop(self):
            """
            Edge property inner product inverse matrix.
            """
            stash_name = f"_MeI_{arg}"
            if getattr(self, stash_name, None) is None:
                prop = getattr(self, arg.lower())
                M_prop = self.mesh.get_edge_inner_product(
                    model=prop, invert_matrix=True
                )
                setattr(self, stash_name, M_prop)
            return getattr(self, stash_name)

        setattr(cls, f"Me{arg}I", MeI_prop)

        def MccDeriv_prop(self, u, v=None, adjoint=False):
            """
            Derivative of `MccProperty` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()
            stash_name = f"_Mcc_{arg}_deriv"

            if getattr(self, stash_name, None) is None:
                M_prop_deriv = sp.diags(self.mesh.cell_volumes) * getattr(
                    self, f"{arg.lower()}Deriv"
                )
                setattr(self, stash_name, M_prop_deriv)
            return __inner_mat_mul_op(
                getattr(self, stash_name), u, v=v, adjoint=adjoint
            )

        setattr(cls, f"Mcc{arg}Deriv", MccDeriv_prop)

        def MnDeriv_prop(self, u, v=None, adjoint=False):
            """
            Derivative of `MnProperty` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()
            stash_name = f"_Mn_{arg}_deriv"
            if getattr(self, stash_name, None) is None:
                M_prop_deriv = (
                    self.mesh.aveN2CC.T
                    * sp.diags(self.mesh.cell_volumes)
                    * getattr(self, f"{arg.lower()}Deriv")
                )
                setattr(self, stash_name, M_prop_deriv)
            return __inner_mat_mul_op(
                getattr(self, stash_name), u, v=v, adjoint=adjoint
            )

        setattr(cls, f"Mn{arg}Deriv", MnDeriv_prop)

        def MfDeriv_prop(self, u, v=None, adjoint=False):
            """
            Derivative of `MfProperty` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()
            stash_name = f"_Mf_{arg}_deriv"
            if getattr(self, stash_name, None) is None:
                M_prop_deriv = self.mesh.get_face_inner_product_deriv(
                    np.ones(self.mesh.n_cells)
                )(np.ones(self.mesh.n_faces)) * getattr(self, f"{arg.lower()}Deriv")
                setattr(self, stash_name, M_prop_deriv)
            return __inner_mat_mul_op(
                getattr(self, stash_name), u, v=v, adjoint=adjoint
            )

        setattr(cls, f"Mf{arg}Deriv", MfDeriv_prop)

        def MeDeriv_prop(self, u, v=None, adjoint=False):
            """
            Derivative of `MeProperty` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()
            stash_name = f"_Me_{arg}_deriv"
            if getattr(self, stash_name, None) is None:
                M_prop_deriv = self.mesh.get_edge_inner_product_deriv(
                    np.ones(self.mesh.n_cells)
                )(np.ones(self.mesh.n_edges)) * getattr(self, f"{arg.lower()}Deriv")
                setattr(self, stash_name, M_prop_deriv)
            return __inner_mat_mul_op(
                getattr(self, stash_name), u, v=v, adjoint=adjoint
            )

        setattr(cls, f"Me{arg}Deriv", MeDeriv_prop)

        def MccIDeriv_prop(self, u, v=None, adjoint=False):
            """
            Derivative of `MccPropertyI` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()

            MI_prop = getattr(self, f"Mcc{arg}I")
            u = MI_prop @ (MI_prop @ -u)
            M_prop_deriv = getattr(self, f"Mcc{arg}Deriv")
            return M_prop_deriv(u, v, adjoint=adjoint)

        setattr(cls, f"Mcc{arg}IDeriv", MccIDeriv_prop)

        def MnIDeriv_prop(self, u, v=None, adjoint=False):
            """
            Derivative of `MnPropertyI` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()

            MI_prop = getattr(self, f"Mn{arg}I")
            u = MI_prop @ (MI_prop @ -u)
            M_prop_deriv = getattr(self, f"Mn{arg}Deriv")
            return M_prop_deriv(u, v, adjoint=adjoint)

        setattr(cls, f"Mn{arg}IDeriv", MnIDeriv_prop)

        def MfIDeriv_prop(self, u, v=None, adjoint=False):
            """I
            Derivative of `MfPropertyI` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()

            MI_prop = getattr(self, f"Mf{arg}I")
            u = MI_prop @ (MI_prop @ -u)
            M_prop_deriv = getattr(self, f"Mf{arg}Deriv")
            return M_prop_deriv(u, v, adjoint=adjoint)

        setattr(cls, f"Mf{arg}IDeriv", MfIDeriv_prop)

        def MeIDeriv_prop(self, u, v=None, adjoint=False):
            """
            Derivative of `MePropertyI` with respect to the model.
            """
            if getattr(self, f"{arg.lower()}Map") is None:
                return Zero()
            if isinstance(u, Zero) or isinstance(v, Zero):
                return Zero()

            MI_prop = getattr(self, f"Me{arg}I")
            u = MI_prop @ (MI_prop @ -u)
            M_prop_deriv = getattr(self, f"Me{arg}Deriv")
            return M_prop_deriv(u, v, adjoint=adjoint)

        setattr(cls, f"Me{arg}IDeriv", MeIDeriv_prop)

        @property
        def _clear_on_prop_update(self):
            items = [
                f"_Mcc_{arg}",
                f"_Mn_{arg}",
                f"_Mf_{arg}",
                f"_Me_{arg}",
                f"_MccI_{arg}",
                f"_MnI_{arg}",
                f"_MfI_{arg}",
                f"_MeI_{arg}",
                f"_Mcc_{arg}_deriv",
                f"_Mn_{arg}_deriv",
                f"_Mf_{arg}_deriv",
                f"_Me_{arg}_deriv",
            ]
            return items

        setattr(cls, f"_clear_on_{arg.lower()}_update", _clear_on_prop_update)
        return cls

    return decorator


class BasePDESimulation(BaseSimulation):
    @property
    def Vol(self):
        return self.Mcc

    @property
    def Mcc(self):
        """
        Cell center inner product matrix.
        """
        if getattr(self, "_Mcc", None) is None:
            self._Mcc = sp.diags(self.mesh.cell_volumes, format="csr")
        return self._Mcc

    @property
    def Mn(self):
        """
        Node inner product matrix.
        """
        if getattr(self, "_Mn", None) is None:
            vol = self.mesh.cell_volumes
            self._Mn = sp.diags(self.mesh.aveN2CC.T * vol, format="csr")
        return self._Mn

    @property
    def Mf(self):
        """
        Face inner product matrix.
        """
        if getattr(self, "_Mf", None) is None:
            self._Mf = self.mesh.get_face_inner_product()
        return self._Mf

    @property
    def Me(self):
        """
        Edge inner product matrix.
        """
        if getattr(self, "_Me", None) is None:
            self._Me = self.mesh.get_edge_inner_product()
        return self._Me

    @property
    def MccI(self):
        if getattr(self, "_MccI", None) is None:
            self._MccI = sp.diags(1.0 / self.mesh.cell_volumes, format="csr")
        return self._MccI

    @property
    def MnI(self):
        """
        Node inner product inverse matrix.
        """
        if getattr(self, "_MnI", None) is None:
            vol = self.mesh.cell_volumes
            self._MnI = sp.diags(1.0 / (self.mesh.aveN2CC.T * vol), format="csr")
        return self._MnI

    @property
    def MfI(self):
        """
        Face inner product inverse matrix.
        """
        if getattr(self, "_MfI", None) is None:
            self._MfI = self.mesh.get_face_inner_product(invert_matrix=True)
        return self._MfI

    @property
    def MeI(self):
        """
        Edge inner product inverse matrix.
        """
        if getattr(self, "_MeI", None) is None:
            self._MeI = self.mesh.get_edge_inner_product(invert_matrix=True)
        return self._MeI


@with_property_mass_matrices("sigma")
@with_property_mass_matrices("rho")
class BaseElectricalPDESimulation(BasePDESimulation):
    sigma, sigmaMap, sigmaDeriv = props.Invertible("Electrical conductivity (S/m)")
    rho, rhoMap, rhoDeriv = props.Invertible("Electrical resistivity (Ohm m)")
    props.Reciprocal(sigma, rho)

    def __init__(
        self, mesh, sigma=None, sigmaMap=None, rho=None, rhoMap=None, **kwargs
    ):
        super().__init__(mesh=mesh, **kwargs)
        self.sigma = sigma
        self.rho = rho
        self.sigmaMap = sigmaMap
        self.rhoMap = rhoMap

    @property
    def deleteTheseOnModelUpdate(self):
        """
        matrices to be deleted if the model for conductivity/resistivity is updated
        """
        toDelete = super().deleteTheseOnModelUpdate
        if self.sigmaMap is not None or self.rhoMap is not None:
            toDelete = (
                toDelete + self._clear_on_sigma_update + self._clear_on_rho_update
            )
        return toDelete

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name in ["sigma", "rho"]:
            for mat in self._clear_on_sigma_update + self._clear_on_rho_update:
                if hasattr(self, mat):
                    delattr(self, mat)


@with_property_mass_matrices("mu")
@with_property_mass_matrices("mui")
class BaseMagneticPDESimulation(BasePDESimulation):
    mu, muMap, muDeriv = props.Invertible(
        "Magnetic Permeability (H/m)",
    )
    mui, muiMap, muiDeriv = props.Invertible("Inverse Magnetic Permeability (m/H)")
    props.Reciprocal(mu, mui)

    def __init__(self, mesh, mu=mu_0, muMap=None, mui=None, muiMap=None, **kwargs):
        super().__init__(mesh=mesh, **kwargs)
        self.mu = mu
        self.mui = mui
        self.muMap = muMap
        self.muiMap = muiMap

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name in ["mu", "mui"]:
            for mat in self._clear_on_mu_update + self._clear_on_mui_update:
                if hasattr(self, mat):
                    delattr(self, mat)

    @property
    def deleteTheseOnModelUpdate(self):
        """
        items to be deleted if the model for Magnetic Permeability is updated
        """
        toDelete = super().deleteTheseOnModelUpdate
        if self.muMap is not None or self.muiMap is not None:
            toDelete = toDelete + self._clear_on_mu_update + self._clear_on_mui_update
        return toDelete



from .mat_utils import mkvc
import numpy as np
from scipy.interpolate import griddata
from scipy.spatial import cKDTree
import scipy.sparse as sp
from discretize.utils import active_from_xyz
import warnings


def surface2ind_topo(mesh, topo, gridLoc="CC", method="nearest", fill_value=np.nan):
    """Get indices of active cells from topography.

    For a mesh and surface topography, this function returns the indices of cells
    lying below the discretized surface topography.

    Parameters
    ----------
    mesh : discretize.TensorMesh or discretize.TreeMesh
        Mesh on which you want to identify active cells
    topo : (n, 3) numpy.ndarray
        Topography data as a ``numpyndarray`` with columns [x,y,z]; can use [x,z] for 2D meshes.
        Topography data can be unstructured.
    gridLoc : str {'CC', 'N'}
        If 'CC', all cells whose centers are below the topography are active cells.
        If 'N', then cells must lie entirely below the topography in order to be active cells.
    method : str {'nearest','linear'}
        Interpolation method for approximating topography at cell's horizontal position.
        Default is 'nearest'.
    fill_value : float
        Defines the elevation for cells outside the horizontal extent of the topography data.
        Default is :py:class:`numpy.nan`.

    Returns
    -------
    (n_active) numpy.ndarray of int
        Indices of active cells below xyz.
    """
    warnings.warn(
        "The surface2ind_topo function has been deprecated, please import "
        "discretize.utils.active_from_xyz. This will be removed in SimPEG 0.20.0",
        FutureWarning,
    )

    active_cells = active_from_xyz(mesh, topo, gridLoc, method)
    return np.arange(mesh.n_cells)[active_cells]


def surface_layer_index(mesh, topo, index=0):
    """Find ith layer of cells below topo for a tensor mesh.

    Parameters
    ----------
    mesh : discretize.TensorMesh
        Input mesh
    topo : (n, 3) numpy.ndarray
        Topography data as a numpy array with columns [x,y,z]; can use [x,z] for 2D meshes.
        Topography data can be unstructured.
    index : int
        How many layers below the surface you want to find

    Returns
    -------
    numpy.ndarray of int
        Index vector for layer of cells
    """

    actv = np.zeros(mesh.nC, dtype="bool")
    # Get cdkTree to find top layer
    tree = cKDTree(mesh.gridCC)

    def ismember(a, b):
        bind = {}
        for i, elt in enumerate(b):
            if elt not in bind:
                bind[elt] = i
        return np.vstack([bind.get(itm, None) for itm in a])

    grid_x, grid_y = np.meshgrid(mesh.cell_centers_x, mesh.cell_centers_y)
    zInterp = mkvc(
        griddata(topo[:, :2], topo[:, 2], (grid_x, grid_y), method="nearest")
    )

    # Get nearest cells
    r, inds = tree.query(np.c_[mkvc(grid_x), mkvc(grid_y), zInterp])
    inds = np.unique(inds)

    # Extract vertical neighbors from Gradz operator
    Dz = mesh.stencil_cell_gradient_z
    Iz, Jz, _ = sp.find(Dz)
    jz = np.sort(Jz[np.argsort(Iz)].reshape((int(Iz.shape[0] / 2), 2)), axis=1)
    for _ in range(index):
        members = ismember(inds, jz[:, 1])
        inds = np.squeeze(jz[members, 0])

    actv[inds] = True

    return actv


def depth_weighting(
    mesh, reference_locs, active_cells=None, exponent=2.0, threshold=None, **kwargs
):
    r"""
    Construct diagonal elements of a depth weighting matrix

    Builds the model weights following the depth weighting strategy, a method
    to generate weights based on the vertical distance between mesh cell
    centers and some reference location(s).
    Use these weights in regularizations to counteract the natural decay of
    potential field data with depth.

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
        Discretized model space.
    reference_locs : float or (n, ndim) numpy.ndarray
        Reference location for the depth weighting.
        It can be a ``float``, which value is the vertical component for
        the reference location.
        Or it can be a 2d array, with multiple reference locations, where each
        row should contain the coordinates of a single location point in the
        following order: _x_, _y_, _z_ (for 3D meshes) or _x_, _z_ (for 2D
        meshes).
        The vertical coordinate of the reference location for each cell in the
        mesh will be obtained by the closest point in ``reference_locs`` using
        only their horizontal coordinates.
    active_cells : (mesh.n_cells) numpy.ndarray of bool, optional
        Index vector for the active cells on the mesh.
        If ``None``, every cell will be assumed to be active.
    exponent : float, optional
        Exponent parameter for depth weighting.
        The exponent should match the natural decay power of the potential
        field. For example, for gravity acceleration, set it to 2; for magnetic
        fields, to 3.
    threshold : float or None, optional
        Threshold parameters used in the depth weighting.
        If ``None``, it will be set to half of the smallest cell width.

    Returns
    -------
    (n_active) numpy.ndarray
        Normalized depth weights for the mesh at every active cell as
        a 1d-array.

    Notes
    -----
    Each diagonal term of the matrix is defined as:

    .. math::

        w(z) = \frac{1}{(|z - z_0| + \epsilon) ^ {\nu / 2}}

    where :math:`z` is the vertical coordinate of the mesh cell centers,
    :math:`z_0` is the vertical coordinate of the reference location,
    :math:`\nu` is the _exponent_,
    and :math:`\epsilon` is a given _threshold_.

    The depth weights array is finally normalized by dividing for its maximum
    value.
    """

    if "indActive" in kwargs:
        warnings.warn(
            "The indActive keyword argument has been deprecated, please use active_cells. "
            "This will be removed in SimPEG 0.19.0",
            FutureWarning,
        )
        active_cells = kwargs["indActive"]

    # Default threshold value
    if threshold is None:
        threshold = 0.5 * mesh.h_gridded.min()

    reference_locs = np.asarray(reference_locs)

    # Calculate depth from receiver locations, delta_z
    # reference_locs is a scalar
    if reference_locs.ndim < 2:
        delta_z = np.abs(mesh.cell_centers[:, -1] - reference_locs)

    # reference_locs is a 2d array
    elif reference_locs.ndim == 2:
        tree = cKDTree(reference_locs[:, :-1])
        _, ind = tree.query(mesh.cell_centers[:, :-1])
        delta_z = np.abs(mesh.cell_centers[:, -1] - reference_locs[ind, -1])

    else:
        raise ValueError("reference_locs must be either a scalar or 2d array!")

    wz = (delta_z + threshold) ** (-0.5 * exponent)

    if active_cells is not None:
        wz = wz[active_cells]

    return wz / np.nanmax(wz)



import numpy as np


from ...survey import BaseSurvey
from .sources import BaseSrcVRM
from ...utils import validate_list_of_types, validate_ndarray_with_shape


############################################
# BASE VRM SURVEY CLASS
############################################


class SurveyVRM(BaseSurvey):
    """Base VRM survey

    Parameters
    ----------
    source_list : list of SimPEG.electromagnetic.viscous_remanent_magnetization.sources.BaseSrcVRM
        List of SimPEG VRM sources
    t_active : numpy.ndarray of bool
        Active time channels used in inversion
    """

    def __init__(self, source_list, t_active=None, **kwargs):
        super(SurveyVRM, self).__init__(source_list=source_list, **kwargs)

        self._nD_all = self.vnD.sum()
        self._nD = self._nD_all

        if t_active is None:
            t_active = np.ones(self._nD_all, dtype=bool)
        self.t_active = validate_ndarray_with_shape(
            "t_active", t_active, shape=(self._nD_all,), dtype=bool
        )

    @property
    def source_list(self):
        """List of VRM sources associated with the survey

        Returns
        -------
        list of BaseSrcVRM
            List of VRM sources associated with the survey
        """
        return self._source_list

    @source_list.setter
    def source_list(self, new_list):
        self._source_list = validate_list_of_types(
            "source_list", new_list, BaseSrcVRM, ensure_unique=True
        )

    @property
    def nD(self):
        if self._nD is None:
            self._nD = np.sum(self.t_active)
        return self._nD

    def set_active_interval(self, tmin, tmax):
        """Set active times using an interval"""

        source_list = self.source_list
        nSrc = len(source_list)
        tActBool = np.array([])

        for pp in range(0, nSrc):
            receiver_list = source_list[pp].receiver_list
            nRx = len(receiver_list)

            for qq in range(0, nRx):
                times = receiver_list[qq].times
                nLoc = np.shape(receiver_list[qq].locations)[0]
                tActBool = np.r_[tActBool, np.kron(np.ones(nLoc), times)]

        self.t_active = (tActBool >= tmin) & (tActBool <= tmax)
        self._nD = np.sum(self.t_active)



import numpy as np
from scipy.constants import epsilon_0

from ...fields import TimeFields
from ...utils import mkvc, sdiag, Zero


class FieldsTDEM(TimeFields):
    r"""
    Fancy Field Storage for a TDEM simulation. Only one field type is stored for
    each problem, the rest are computed. The fields obejct acts like an array
    and is indexed by

    .. code-block:: python

        f = problem.fields(m)
        e = f[source_list,'e']
        b = f[source_list,'b']

    If accessing all sources for a given field, use the :code:`:`

    .. code-block:: python

        f = problem.fields(m)
        e = f[:,'e']
        b = f[:,'b']

    The array returned will be size (nE or nF, nSrcs :math:`\times`
    nFrequencies)
    """

    knownFields = {}
    dtype = float

    def _GLoc(self, fieldType):
        """Grid location of the fieldType"""
        return self.aliasFields[fieldType][1]

    def _eDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._eDeriv_u(tInd, src, v, adjoint),
                self._eDeriv_m(tInd, src, v, adjoint),
            )
        return self._eDeriv_u(tInd, src, dun_dm_v) + self._eDeriv_m(tInd, src, v)

    def _bDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._bDeriv_u(tInd, src, v, adjoint),
                self._bDeriv_m(tInd, src, v, adjoint),
            )
        return self._bDeriv_u(tInd, src, dun_dm_v) + self._bDeriv_m(tInd, src, v)

    def _dbdtDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._dbdtDeriv_u(tInd, src, v, adjoint),
                Zero()
                # self._dbdtDeriv_m(tInd, src, v, adjoint),
            )
        return self._dbdtDeriv_u(tInd, src, dun_dm_v) + self._dbdtDeriv_m(tInd, src, v)

    def _hDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._hDeriv_u(tInd, src, v, adjoint),
                self._hDeriv_m(tInd, src, v, adjoint),
            )
        return self._hDeriv_u(tInd, src, dun_dm_v) + self._hDeriv_m(tInd, src, v)

    def _dhdtDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._dhdtDeriv_u(tInd, src, v, adjoint),
                self._dhdtDeriv_m(tInd, src, v, adjoint),
            )
        return self._dhdtDeriv_u(tInd, src, dun_dm_v) + self._dhdtDeriv_m(tInd, src, v)

    def _jDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._jDeriv_u(tInd, src, v, adjoint),
                self._jDeriv_m(tInd, src, v, adjoint),
            )
        return self._jDeriv_u(tInd, src, dun_dm_v) + self._jDeriv_m(tInd, src, v)


class FieldsDerivativesEB(FieldsTDEM):
    """
    A fields object for satshing derivs in the EB formulation
    """

    knownFields = {
        "bDeriv": "F",
        "eDeriv": "E",
        "hDeriv": "F",
        "jDeriv": "E",
        "dbdtDeriv": "F",
        "dhdtDeriv": "F",
    }


class FieldsDerivativesHJ(FieldsTDEM):
    """
    A fields object for satshing derivs in the HJ formulation
    """

    knownFields = {
        "bDeriv": "E",
        "eDeriv": "F",
        "hDeriv": "E",
        "jDeriv": "F",
        "dbdtDeriv": "E",
        "dhdtDeriv": "E",
    }


class Fields3DMagneticFluxDensity(FieldsTDEM):
    """Field Storage for a TDEM simulation."""

    knownFields = {"bSolution": "F"}
    aliasFields = {
        "b": ["bSolution", "F", "_b"],
        "h": ["bSolution", "F", "_h"],
        "e": ["bSolution", "E", "_e"],
        "j": ["bSolution", "E", "_j"],
        "dbdt": ["bSolution", "F", "_dbdt"],
        "dhdt": ["bSolution", "F", "_dhdt"],
    }

    def startup(self):
        self._times = self.simulation.times
        self._MeSigma = self.simulation.MeSigma
        self._MeSigmaI = self.simulation.MeSigmaI
        self._MeSigmaDeriv = self.simulation.MeSigmaDeriv
        self._MeSigmaIDeriv = self.simulation.MeSigmaIDeriv
        self._edgeCurl = self.simulation.mesh.edge_curl
        self._MfMui = self.simulation.MfMui
        self._timeMesh = self.simulation.time_mesh

    def _TLoc(self, fieldType):
        return "N"

    def _b(self, bSolution, source_list, tInd):
        return bSolution

    def _bDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _bDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dbdt(self, bSolution, source_list, tInd):
        # self._timeMesh.face_divergence
        dbdt = -self._edgeCurl * self._e(bSolution, source_list, tInd)
        for i, src in enumerate(source_list):
            s_m = src.s_m(self.simulation, self._times[tInd])
            dbdt[:, i] = dbdt[:, i] + s_m
        return dbdt

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return -self._eDeriv_u(tInd, src, self._edgeCurl.T @ dun_dm_v, adjoint)
        return -(self._edgeCurl * self._eDeriv_u(tInd, src, dun_dm_v))

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint is True:
            return -(self._eDeriv_m(tInd, src, self._edgeCurl.T * v, adjoint))
        return -(
            self._edgeCurl * self._eDeriv_m(tInd, src, v)
        )  # + src.s_mDeriv() assuming src doesn't have deriv for now

    def _e(self, bSolution, source_list, tInd):
        e = self._MeSigmaI * (self._edgeCurl.T * (self._MfMui * bSolution))
        for i, src in enumerate(source_list):
            s_e = src.s_e(self.simulation, self._times[tInd])
            e[:, i] = e[:, i] - self._MeSigmaI * s_e
        return e

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return self._MfMui.T @ (self._edgeCurl @ (self._MeSigmaI.T @ dun_dm_v))
        return self._MeSigmaI * (self._edgeCurl.T * (self._MfMui * dun_dm_v))

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        _, s_e = src.eval(self.simulation, self._times[tInd])
        bSolution = self[[src], "bSolution", tInd].flatten()

        _, s_eDeriv = src.evalDeriv(self._times[tInd], self, adjoint=adjoint)

        if adjoint is True:
            return self._MeSigmaIDeriv(
                -s_e + self._edgeCurl.T * (self._MfMui * bSolution), v, adjoint
            ) - s_eDeriv(self._MeSigmaI.T * v)

        return self._MeSigmaIDeriv(
            -s_e + self._edgeCurl.T * (self._MfMui * bSolution), v, adjoint
        ) - self._MeSigmaI * s_eDeriv(v)

    def _j(self, hSolution, source_list, tInd):
        return self.simulation.MeI * (
            self._MeSigma * self._e(hSolution, source_list, tInd)
        )

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._eDeriv_u(
                tInd,
                src,
                self._MeSigma.T * (self.simulation.MeI.T * dun_dm_v),
                adjoint=True,
            )
        return self.simulation.MeI * (
            self._MeSigma * self._eDeriv_u(tInd, src, dun_dm_v)
        )

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        e = self[src, "e", tInd]
        if adjoint:
            w = self.simulation.MeI.T * v
            return self._MeSigmaDeriv(e).T * w + self._eDeriv_m(
                tInd, src, self._MeSigma.T * w, adjoint=True
            )
        return self.simulation.MeI * (
            self._MeSigmaDeriv(e) * v + self._MeSigma * self._eDeriv_m(tInd, src, v)
        )

    def _h(self, hSolution, source_list, tInd):
        return self.simulation.MfI * (
            self._MfMui * self._b(hSolution, source_list, tInd)
        )

    def _hDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._bDeriv_u(
                tInd,
                src,
                self._MfMui.T * (self.simulation.MfI.T * dun_dm_v),
                adjoint=True,
            )
        return self.simulation.MfI * (self._MfMui * self._bDeriv_u(tInd, src, dun_dm_v))

    def _hDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._bDeriv_m(
                tInd, src, self._MfMui.T * (self.simulation.MfI.T * v), adjoint=True
            )
        return self.simulation.MfI * (self._MfMui * self._bDeriv_m(tInd, src, v))

    def _dhdt(self, hSolution, source_list, tInd):
        return self.simulation.MfI * (
            self._MfMui * self._dbdt(hSolution, source_list, tInd)
        )

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_u(
                tInd,
                src,
                self._MfMui.T * (self.simulation.MfI.T * dun_dm_v),
                adjoint=True,
            )
        return self.simulation.MfI * (
            self._MfMui * self._dbdtDeriv_u(tInd, src, dun_dm_v)
        )

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_m(
                tInd, src, self._MfMui.T * (self.simulation.MfI.T * v), adjoint=True
            )
        return self.simulation.MfI * (self._MfMui * self._dbdtDeriv_m(tInd, src, v))


class Fields3DElectricField(FieldsTDEM):
    """Fancy Field Storage for a TDEM simulation."""

    knownFields = {"eSolution": "E"}
    aliasFields = {
        "e": ["eSolution", "E", "_e"],
        "j": ["eSolution", "E", "_j"],
        "b": ["eSolution", "F", "_b"],
        # 'h': ['eSolution', 'F', '_h'],
        "dbdt": ["eSolution", "F", "_dbdt"],
        "dhdt": ["eSolution", "F", "_dhdt"],
    }

    def startup(self):
        self._times = self.simulation.times
        self._MeSigma = self.simulation.MeSigma
        self._MeSigmaI = self.simulation.MeSigmaI
        self._MeSigmaDeriv = self.simulation.MeSigmaDeriv
        self._MeSigmaIDeriv = self.simulation.MeSigmaIDeriv
        self._edgeCurl = self.simulation.mesh.edge_curl
        self._MfMui = self.simulation.MfMui

    def _TLoc(self, fieldType):
        return "N"

    def _e(self, eSolution, source_list, tInd):
        return eSolution

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dbdt(self, eSolution, source_list, tInd):
        s_m = np.zeros((self.mesh.nF, len(source_list)))
        for i, src in enumerate(source_list):
            s_m_src = src.s_m(self.simulation, self._times[tInd])
            s_m[:, i] = s_m[:, i] + s_m_src
        return s_m - self._edgeCurl * eSolution

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return -self._edgeCurl.T * dun_dm_v
        return -self._edgeCurl * dun_dm_v

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        # s_mDeriv = src.s_mDeriv(
        #     self._times[tInd], self, adjoint=adjoint
        # )
        return Zero()  # assumes source doesn't depend on model

    def _b(self, eSolution, source_list, tInd):
        """
        Integrate _db_dt using rectangles
        """
        raise NotImplementedError(
            "To obtain b-fields, please use Simulation3DMagneticFluxDensity"
        )
        # dbdt = self._dbdt(eSolution, source_list, tInd)
        # dt = self.simulation.time_mesh.h[0]
        # # assume widths of "ghost cells" same on either end
        # dtn = np.hstack([dt[0], 0.5*(dt[1:] + dt[:-1]), dt[-1]])
        # return dtn[tInd] * dbdt
        # # raise NotImplementedError

    def _j(self, eSolution, source_list, tInd):
        return self.simulation.MeI * (
            self._MeSigma * self._e(eSolution, source_list, tInd)
        )

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._eDeriv_u(
                tInd,
                src,
                self._MeSigma.T * (self.simulation.MeI.T * dun_dm_v),
                adjoint=True,
            )
        return self.simulation.MeI * (
            self._MeSigma * self._eDeriv_u(tInd, src, dun_dm_v)
        )

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        e = self[src, "e", tInd]
        if adjoint:
            w = self.simulation.MeI.T * v
            return self._MeSigmaDeriv(e).T * w + self._eDeriv_m(
                tInd, src, self._MeSigma.T * w, adjoint=True
            )
        return self.simulation.MeI * (
            self._MeSigmaDeriv(e) * v + self._MeSigma * self._eDeriv_m(tInd, src, v)
        )

    def _dhdt(self, eSolution, source_list, tInd):
        return self.simulation.MfI * (
            self._MfMui * self._dbdt(eSolution, source_list, tInd)
        )

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_u(
                tInd,
                src,
                self._MfMui.T * (self.simulation.MfI.T * dun_dm_v),
                adjoint=True,
            )
        return self.simulation.MfI * (
            self._MfMui * self._dbdtDeriv_u(tInd, src, dun_dm_v)
        )

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_m(
                tInd, src, self._MfMui.T * (self.simulation.MfI.T * v)
            )
        return self.simulation.MfI * (self._MfMui * self._dbdtDeriv_m(tInd, src, v))


class Fields3DMagneticField(FieldsTDEM):
    """Fancy Field Storage for a TDEM simulation."""

    knownFields = {"hSolution": "E"}
    aliasFields = {
        "h": ["hSolution", "E", "_h"],
        "b": ["hSolution", "E", "_b"],
        "dhdt": ["hSolution", "E", "_dhdt"],
        "dbdt": ["hSolution", "E", "_dbdt"],
        "j": ["hSolution", "F", "_j"],
        "e": ["hSolution", "F", "_e"],
        "charge": ["hSolution", "CC", "_charge"],
    }

    def startup(self):
        self._times = self.simulation.times
        self._edgeCurl = self.simulation.mesh.edge_curl
        self._MeMuI = self.simulation.MeMuI
        self._MeMu = self.simulation.MeMu
        self._MfRho = self.simulation.MfRho
        self._MfRhoDeriv = self.simulation.MfRhoDeriv

    def _TLoc(self, fieldType):
        # if fieldType in ['h', 'j']:
        return "N"
        # else:
        #     raise NotImplementedError

    def _h(self, hSolution, source_list, tInd):
        return hSolution

    def _hDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _hDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dhdt(self, hSolution, source_list, tInd):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho

        dhdt = -MeMuI * (C.T * (MfRho * (C * hSolution)))

        for i, src in enumerate(source_list):
            s_m, s_e = src.eval(self.simulation, self._times[tInd])
            dhdt[:, i] = MeMuI * (C.T * MfRho * s_e + s_m) + dhdt[:, i]
        return dhdt

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho

        if adjoint:
            return -C.T * (MfRho.T * (C * (MeMuI * dun_dm_v)))
        return -MeMuI * (C.T * (MfRho * (C * dun_dm_v)))

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRhoDeriv = self._MfRhoDeriv

        hSolution = self[[src], "hSolution", tInd].flatten()
        s_e = src.s_e(self.simulation, self._times[tInd])

        if adjoint:
            return -MfRhoDeriv(C * hSolution - s_e, (C * (MeMuI * v)), adjoint)
        return -MeMuI * (C.T * (MfRhoDeriv(C * hSolution - s_e, v, adjoint)))

    def _j(self, hSolution, source_list, tInd):
        s_e = np.zeros((self.mesh.nF, len(source_list)))
        for i, src in enumerate(source_list):
            s_e_src = src.s_e(self.simulation, self._times[tInd])
            s_e[:, i] = s_e[:, i] + s_e_src

        return self._edgeCurl * hSolution - s_e

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._edgeCurl.T * dun_dm_v
        return self._edgeCurl * dun_dm_v

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()  # assumes the source doesn't depend on the model

    def _b(self, hSolution, source_list, tInd):
        h = self._h(hSolution, source_list, tInd)
        return self.simulation.MeI * (self._MeMu * h)

    def _bDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._hDeriv_u(
                tInd,
                src,
                self._MeMu.T * (self.simulation.MeI.T * dun_dm_v),
                adjoint=adjoint,
            )
        return self.simulation.MeI * (self._MeMu * self._hDeriv_u(tInd, src, dun_dm_v))

    def _bDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._hDeriv_m(
                tInd, src, self._MeMu.T * (self.simulation.MeI.T * v), adjoint=adjoint
            )
        return self.simulation.MeI * (self._MeMu * self._hDeriv_m(tInd, src, v))

    def _dbdt(self, hSolution, source_list, tInd):
        dhdt = self._dhdt(hSolution, source_list, tInd)
        return self.simulation.MeI * (self._MeMu * dhdt)

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._dhdtDeriv_u(
                tInd,
                src,
                self._MeMu.T * (self.simulation.MeI.T * dun_dm_v),
                adjoint=adjoint,
            )
        return self.simulation.MeI * (
            self._MeMu * self._dhdtDeriv_u(tInd, src, dun_dm_v)
        )

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dhdtDeriv_m(
                tInd, src, self._MeMu.T * (self.simulation.MeI.T * v), adjoint=adjoint
            )
        return self.simulation.MeI * (self._MeMu * self._dhdtDeriv_m(tInd, src, v))

    def _e(self, hSolution, source_list, tInd):
        return self.simulation.MfI * (
            self._MfRho * self._j(hSolution, source_list, tInd)
        )

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._jDeriv_u(
                tInd,
                src,
                self._MfRho.T * (self.simulation.MfI.T * dun_dm_v),
                adjoint=True,
            )
        return self.simulation.MfI * (self._MfRho * self._jDeriv_u(tInd, src, dun_dm_v))

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        j = mkvc(self[src, "j", tInd])
        if adjoint is True:
            return self._MfRhoDeriv(
                j, self.simulation.MfI.T * v, adjoint
            ) + self._jDeriv_m(tInd, src, self._MfRho * v)
        return self.simulation.MfI * (
            self._MfRhoDeriv(j, v) + self._MfRho * self._jDeriv_m(tInd, src, v)
        )

    def _charge(self, hSolution, source_list, tInd):
        vol = sdiag(self.simulation.mesh.cell_volumes)
        return (
            epsilon_0
            * vol
            * (
                self.simulation.mesh.face_divergence
                * self._e(hSolution, source_list, tInd)
            )
        )


class Fields3DCurrentDensity(FieldsTDEM):
    """Fancy Field Storage for a TDEM simulation."""

    knownFields = {"jSolution": "F"}
    aliasFields = {
        "dhdt": ["jSolution", "E", "_dhdt"],
        "dbdt": ["jSolution", "E", "_dbdt"],
        "j": ["jSolution", "F", "_j"],
        "e": ["jSolution", "F", "_e"],
        "charge": ["jSolution", "CC", "_charge"],
        "charge_density": ["jSolution", "CC", "_charge_density"],
    }

    def startup(self):
        self._times = self.simulation.times
        self._edgeCurl = self.simulation.mesh.edge_curl
        self._MeMuI = self.simulation.MeMuI
        self._MfRho = self.simulation.MfRho
        self._MfRhoDeriv = self.simulation.MfRhoDeriv

    def _TLoc(self, fieldType):
        # if fieldType in ['h', 'j']:
        return "N"

    def _j(self, jSolution, source_list, tInd):
        return jSolution

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _h(self, jSolution, source_list, tInd):
        raise NotImplementedError(
            "Please use Simulation3DMagneticField to get h-fields"
        )

    def _dhdt(self, jSolution, source_list, tInd):
        C = self._edgeCurl
        MfRho = self._MfRho
        MeMuI = self._MeMuI

        dhdt = -MeMuI * (C.T * (MfRho * jSolution))
        for i, src in enumerate(source_list):
            s_m = src.s_m(self.simulation, self.simulation.times[tInd])
            dhdt[:, i] = MeMuI * s_m + dhdt[:, i]
        return dhdt

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        C = self._edgeCurl
        MfRho = self._MfRho
        MeMuI = self._MeMuI

        if adjoint is True:
            return -MfRho.T * (C * (MeMuI.T * dun_dm_v))
        return -MeMuI * (C.T * (MfRho * dun_dm_v))

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        jSolution = self[[src], "jSolution", tInd].flatten()
        C = self._edgeCurl
        MeMuI = self._MeMuI

        if adjoint is True:
            return -self._MfRhoDeriv(jSolution, C * (MeMuI * v), adjoint)
        return -MeMuI * (C.T * (self._MfRhoDeriv(jSolution, v)))

    def _e(self, jSolution, source_list, tInd):
        return self.simulation.MfI * (
            self._MfRho * self._j(jSolution, source_list, tInd)
        )

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return self._MfRho.T * (self.simulation.MfI.T * dun_dm_v)
        return self.simulation.MfI * (self._MfRho * dun_dm_v)

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        jSolution = mkvc(self[src, "jSolution", tInd])
        if adjoint:
            return self._MfRhoDeriv(jSolution, self.simulation.MfI.T * v, adjoint)
        return self.simulation.MfI * self._MfRhoDeriv(jSolution, v)

    def _charge(self, jSolution, source_list, tInd):
        vol = sdiag(self.simulation.mesh.cell_volumes)
        return vol * self._charge_density(jSolution, source_list, tInd)

    def _charge_density(self, jSolution, source_list, tInd):
        return epsilon_0 * (
            self.simulation.mesh.face_divergence * self._e(jSolution, source_list, tInd)
        )

    def _dbdt(self, jSolution, source_list, tInd):
        dhdt = mkvc(self._dhdt(jSolution, source_list, tInd))
        return self.simulation.MeI * (self.simulation.MeMu * dhdt)

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        # dhdt = mkvc(self[src, 'dhdt', tInd])
        if adjoint:
            return self._dhdtDeriv_u(
                tInd,
                src,
                self.simulation.MeMu.T * (self.simulation.MeI.T * dun_dm_v),
                adjoint,
            )
        return self.simulation.MeI * (
            self.simulation.MeMu * self._dhdtDeriv_u(tInd, src, dun_dm_v)
        )

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dhdtDeriv_m(
                tInd, src, self.simulation.MeMu.T * (self.simulation.MeI.T * v), adjoint
            )
        return self.simulation.MeI * (
            self.simulation.MeMu * self._dhdtDeriv_m(tInd, src, v)
        )



import numpy as np

from .... import survey
from ....utils import Zero, mkvc, validate_list_of_types, validate_float
from .receivers import BaseRx


class BaseSrc(survey.BaseSrc):
    """Base spectral IP source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.static.resistivity.receivers.BaseRx
        A list of DC/IP receivers
    location : (dim) numpy.ndarray
        Source location
    current : float, default=1.0
        Current amplitude [A]
    """

    def __init__(self, receiver_list, location, current=1.0, **kwargs):
        super(BaseSrc, self).__init__(
            receiver_list=receiver_list, location=location, **kwargs
        )
        self.current = current

    @property
    def receiver_list(self):
        """List of receivers associated with the source

        Returns
        -------
        list of SimPEG.electromagnetics.static.spectral_induced_polarization.receivers.BaseRx
            List of receivers associated with the source
        """
        return self._receiver_list

    @receiver_list.setter
    def receiver_list(self, new_list):
        self._receiver_list = validate_list_of_types(
            "source_list", new_list, BaseRx, ensure_unique=True
        )

    @property
    def current(self):
        """Source current

        Returns
        -------
        float
            Source current
        """
        return self._current

    @current.setter
    def current(self, I):
        I = validate_float("current", I)
        if I == 0.0:
            raise ValueError("current must be non-zero.")
        self._current = I

    def eval(self, simulation):  # noqa: A003
        """Not implemented for BaseSrc"""
        raise NotImplementedError

    def evalDeriv(self, simulation):
        """Returns zero"""
        return Zero()

    @property
    def nD(self):
        """Number of data

        Returns
        -------
        int
            Number of data
        """
        return self.vnD.sum()

    @property
    def vnD(self):
        """Vector number of data

        Returns
        -------
        list of int
            Number of data per source
        """
        return np.array([rx.nD * len(rx.times) for rx in self.receiver_list])


class Dipole(BaseSrc):
    """Spectral IP dipole source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.static.spectral_induced_polarization.receivers.BaseRx
        A list of spectral IP receivers
    location_a : (dim) numpy.ndarray
        A electrode location; remember to set 'location_b' keyword argument to define N electrode locations.
    location_b : (dim) numpy.ndarray
        B electrode location; remember to set 'location_a' keyword argument to define M electrode locations.
    location : list or tuple of length 2 of numpy.ndarray
        A and B electrode locations. In this case, do not set the 'location_a' and 'location_b'
        keyword arguments. And we supply a list or tuple of the form [location_a, location_b].
    current : float, default=1.0
        Current amplitude [A]
    """

    def __init__(
        self,
        receiver_list=None,
        location_a=None,
        location_b=None,
        location=None,
        **kwargs,
    ):
        # if location_a set, then use location_a, location_b
        if location_a is not None:
            if location_b is None:
                raise ValueError(
                    "For a dipole source both location_a and location_b " "must be set"
                )

            if location is not None:
                raise ValueError(
                    "Cannot set both location and location_a, location_b. "
                    "Please provide either location=(location_a, location_b) "
                    "or both location_a=location_a, location_b=location_b"
                )

            location = [location_a, location_b]

        if location is None:
            raise AttributeError(
                "Source cannot be instantiated without assigning 'location'."
                "Please provide either location=(location_a, location_b) "
                "or both location_a=location_a, location_b=location_b"
            )

        super(Dipole, self).__init__(receiver_list, location=location, **kwargs)

    # @property
    # def location_a(self):
    #     """Location of the A-electrode"""
    #     return self.location[0]

    # @property
    # def location_b(self):
    #     """Location of the B-electrode"""
    #     return self.location[1]

    @property
    def location(self):
        """A and N electrode locations

        Returns
        -------
        list of np.ndarray of the form [dim, dim]
            A and B electrode locations.
        """
        return self._locations

    @location.setter
    def location(self, locs):
        if len(locs) != 2:
            raise ValueError(
                "locations must be a list or tuple of length 2: "
                "[location_a, location_b. The input locations has "
                f"length {len(locs)}"
            )

        locs = [np.atleast_1d(locs[0]), np.atleast_1d(locs[1])]

        # check the size of locations_m, locations_n
        if locs[0].shape != locs[1].shape:
            raise ValueError(
                f"location_a (shape: {locs[0].shape}) and "
                f"location_b (shape: {locs[1].shape}) need to be "
                f"the same size"
            )

        self._locations = locs

    @property
    def location_a(self):
        """Location of the A-electrode

        Returns
        -------
        (dim) numpy.ndarray
            Location of the A-electrode
        """
        return self.location[0]

    @property
    def location_b(self):
        """Location of the B-electrode

        Returns
        -------
        (dim) numpy.ndarray
            Location of the B-electrode
        """
        return self.location[1]

    def eval(self, simulation):  # noqa: A003
        """Project source to mesh.

        Parameters
        ----------
        sim : SimPEG.electromagnetics.static.spectral_induced_polarization.simulation.BaseDCSimulation
            A spectral IP simulation

        Returns
        -------
        numpy.ndarray
            Discretize source term on the mesh
        """
        if simulation._formulation == "HJ":
            inds = simulation.mesh.closest_points_index(self.location, grid_loc="CC")
            q = np.zeros(simulation.mesh.nC)
            q[inds] = self.current * np.r_[1.0, -1.0]
        elif simulation._formulation == "EB":
            qa = simulation.mesh.get_interpolation_matrix(
                self.location[0], location_type="N"
            ).todense()
            qb = -simulation.mesh.get_interpolation_matrix(
                self.location[1], location_type="N"
            ).todense()
            q = self.current * mkvc(qa + qb)
        return q


class Pole(BaseSrc):
    """Spectral IP pole source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.static.spectral_induced_polarization.receivers.BaseRx
        list of spectral IP receivers
    location : (dim) array_like
        Electrode location
    current : float, default=1.0
        Current amplitude [A]
    """

    def __init__(self, receiver_list=None, location=None, **kwargs):
        super(Pole, self).__init__(receiver_list, location=location, **kwargs)

    def eval(self, simulation):  # noqa: A003
        """Project source to mesh.

        Parameters
        ----------
        sim : SimPEG.electromagnetics.static.resistivity.simulation.BaseDCSimulation
            A DC/IP simulation

        Returns
        -------
        numpy.ndarray
            Discretize source term on the mesh
        """
        if simulation._formulation == "HJ":
            inds = simulation.mesh.closest_points_index(self.location)
            q = np.zeros(simulation.mesh.nC)
            q[inds] = self.current * np.r_[1.0]
        elif simulation._formulation == "EB":
            q = simulation.mesh.get_interpolation_matrix(
                self.location, location_type="N"
            ).todense()
            q = self.current * mkvc(q)
        return q



import numpy as np
from ..regularization import BaseSimilarityMeasure
from ..utils import eigenvalue_by_power_iteration
from ..optimization import IterationPrinters, StoppingCriteria
from .directives import InversionDirective, SaveEveryIteration


###############################################################################
#                                                                             #
#              Directives of joint inversion                                  #
#                                                                             #
###############################################################################
class SimilarityMeasureInversionPrinters:
    betas = {
        "title": "betas",
        "value": lambda M: ["{:.2e}".format(elem) for elem in M.parent.betas],
        "width": 26,
        "format": "%s",
    }
    lambd = {
        "title": "lambda",
        "value": lambda M: M.parent.lambd,
        "width": 10,
        "format": "%1.2e",
    }
    phi_d_list = {
        "title": "phi_d",
        "value": lambda M: ["{:.2e}".format(elem) for elem in M.parent.phi_d_list],
        "width": 26,
        "format": "%s",
    }
    phi_m_list = {
        "title": "phi_m",
        "value": lambda M: ["{:.2e}".format(elem) for elem in M.parent.phi_m_list],
        "width": 26,
        "format": "%s",
    }
    phi_sim = {
        "title": "phi_sim",
        "value": lambda M: M.parent.phi_sim,
        "width": 10,
        "format": "%1.2e",
    }
    iterationCG = {
        "title": "iterCG",
        "value": lambda M: M.cg_count,
        "width": 10,
        "format": "%3d",
    }


class SimilarityMeasureInversionDirective(InversionDirective):
    """
    Directive for two model similiraty measure joint inversions. Sets Printers and
    StoppingCriteria.

    Notes
    -----
    Methods assume we are working with two models, and a single similarity measure.
    Also, the SimilarityMeasure objective function must be the last regularization.
    """

    printers = [
        IterationPrinters.iteration,
        SimilarityMeasureInversionPrinters.betas,
        SimilarityMeasureInversionPrinters.lambd,
        IterationPrinters.f,
        SimilarityMeasureInversionPrinters.phi_d_list,
        SimilarityMeasureInversionPrinters.phi_m_list,
        SimilarityMeasureInversionPrinters.phi_sim,
        SimilarityMeasureInversionPrinters.iterationCG,
    ]

    def initialize(self):
        if not isinstance(self.reg.objfcts[-1], BaseSimilarityMeasure):
            raise TypeError(
                f"The last regularization function must be an instance of "
                f"BaseSimilarityMeasure, got {type(self.reg.objfcts[-1])}."
            )

        # define relevant attributes
        self.betas = self.reg.multipliers[:-1]
        self.lambd = self.reg.multipliers[-1]
        self.phi_d_list = []
        self.phi_m_list = []
        self.phi_sim = 0.0

        # pass attributes to invProb
        self.invProb.betas = self.betas
        self.invProb.num_models = len(self.betas)
        self.invProb.lambd = self.lambd
        self.invProb.phi_d_list = self.phi_d_list
        self.invProb.phi_m_list = self.phi_m_list
        self.invProb.phi_sim = self.phi_sim

        self.opt.printers = self.printers
        self.opt.stoppers = [StoppingCriteria.iteration]

    def validate(self, directiveList):
        # check that this directive is first in the DirectiveList
        dList = directiveList.dList
        self_ind = dList.index(self)
        if self_ind != 0:
            raise IndexError(
                "The CrossGradientInversionDirective must be first in directive list."
            )
        return True

    def endIter(self):
        # compute attribute values
        phi_d = []
        for dmis in self.dmisfit.objfcts:
            phi_d.append(dmis(self.opt.xc))

        phi_m = []
        for reg in self.reg.objfcts:
            phi_m.append(reg(self.opt.xc))

        # pass attributes values to invProb
        self.invProb.phi_d_list = phi_d
        self.invProb.phi_m_list = phi_m[:-1]
        self.invProb.phi_sim = phi_m[-1]
        self.invProb.betas = self.reg.multipliers[:-1]
        # Assume last reg.objfct is the coupling
        self.invProb.lambd = self.reg.multipliers[-1]


class SimilarityMeasureSaveOutputEveryIteration(SaveEveryIteration):
    """
    SaveOutputEveryIteration for Joint Inversions.
    Saves information on the tradeoff parameters, data misfits, regularizations,
    coupling term, number of CG iterations, and value of cost function.
    """

    header = None
    save_txt = True
    betas = None
    phi_d = None
    phi_m = None
    phi_sim = None
    phi = None

    def initialize(self):
        if self.save_txt is True:
            print(
                "CrossGradientSaveOutputEveryIteration will save your inversion "
                "progress as: '###-{0!s}.txt'".format(self.fileName)
            )
            f = open(self.fileName + ".txt", "w")
            self.header = "  #          betas            lambda         joint_phi_d                joint_phi_m            phi_sim       iterCG     phi    \n"
            f.write(self.header)
            f.close()

        # Create a list of each
        self.betas = []
        self.lambd = []
        self.phi_d = []
        self.phi_m = []
        self.phi = []
        self.phi_sim = []

    def endIter(self):
        self.betas.append(["{:.2e}".format(elem) for elem in self.invProb.betas])
        self.phi_d.append(["{:.3e}".format(elem) for elem in self.invProb.phi_d_list])
        self.phi_m.append(["{:.3e}".format(elem) for elem in self.invProb.phi_m_list])
        self.lambd.append("{:.2e}".format(self.invProb.lambd))
        self.phi_sim.append(self.invProb.phi_sim)
        self.phi.append(self.opt.f)

        if self.save_txt:
            f = open(self.fileName + ".txt", "a")
            i = self.opt.iter
            f.write(
                " {0:2d}  {1}  {2}  {3}  {4}  {5:1.4e}  {6:d}  {7:1.4e}\n".format(
                    i,
                    self.betas[i - 1],
                    self.lambd[i - 1],
                    self.phi_d[i - 1],
                    self.phi_m[i - 1],
                    self.phi_sim[i - 1],
                    self.opt.cg_count,
                    self.phi[i - 1],
                )
            )
            f.close()

    def load_results(self):
        results = np.loadtxt(self.fileName + str(".txt"), comments="#")
        self.betas = results[:, 1]
        self.lambd = results[:, 2]
        self.phi_d = results[:, 3]
        self.phi_m = results[:, 4]
        self.phi_sim = results[:, 5]
        self.f = results[:, 7]


class PairedBetaEstimate_ByEig(InversionDirective):
    """
    Estimate the trade-off parameter, beta, between pairs of data misfit(s) and the
    regularization(s) as a multiple of the ratio between the highest eigenvalue of the
    data misfit term and the highest eigenvalue of the regularization.
    The highest eigenvalues are estimated through power iterations and Rayleigh
    quotient.

    Notes
    -----
    This class assumes the order of the data misfits for each model parameter match
    the order for the respective regularizations, i.e.

    >>> data_misfits = [phi_d_m1, phi_d_m2, phi_d_m3]
    >>> regs = [phi_m_m1, phi_m_m2, phi_m_m3]

    In which case it will estimate regularization parameters for each respective pair.
    """

    beta0_ratio = 1.0  #: the estimated ratio is multiplied by this to obtain beta
    n_pw_iter = 4  #: number of power iterations for estimation.
    seed = None  #: Random seed for the directive

    def initialize(self):
        r"""
        The initial beta is calculated by comparing the estimated
        eigenvalues of :math:`J^T J` and :math:`W^T W`.
        To estimate the eigenvector of **A**, we will use one iteration
        of the *Power Method*:

        .. math::

            \mathbf{x_1 = A x_0}

        Given this (very course) approximation of the eigenvector, we can
        use the *Rayleigh quotient* to approximate the largest eigenvalue.

        .. math::

            \lambda_0 = \frac{\mathbf{x^\top A x}}{\mathbf{x^\top x}}

        We will approximate the largest eigenvalue for both JtJ and WtW,
        and use some ratio of the quotient to estimate beta0.

        .. math::

            \beta_0 = \gamma \frac{\mathbf{x^\top J^\top J x}}{\mathbf{x^\top W^\top W x}}

        :rtype: float
        :return: beta0
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        if self.debug:
            print("Calculating the beta0 parameter.")

        m = self.invProb.model
        dmis_eigenvalues = []
        reg_eigenvalues = []
        dmis_objs = self.dmisfit.objfcts
        reg_objs = [
            obj
            for obj in self.reg.objfcts
            if not isinstance(obj, BaseSimilarityMeasure)
        ]
        if len(dmis_objs) != len(reg_objs):
            raise ValueError(
                f"There must be the same number of data misfit and regularizations."
                f"Got {len(dmis_objs)} and {len(reg_objs)} respectively."
            )
        for dmis, reg in zip(dmis_objs, reg_objs):
            dmis_eigenvalues.append(
                eigenvalue_by_power_iteration(
                    dmis,
                    m,
                    n_pw_iter=self.n_pw_iter,
                )
            )

            reg_eigenvalues.append(
                eigenvalue_by_power_iteration(
                    reg,
                    m,
                    n_pw_iter=self.n_pw_iter,
                )
            )

        self.ratios = np.array(dmis_eigenvalues) / np.array(reg_eigenvalues)
        self.invProb.betas = self.beta0_ratio * self.ratios
        self.reg.multipliers[:-1] = self.invProb.betas


class PairedBetaSchedule(InversionDirective):
    """
    Directive for beta cooling schedule to determine the tradeoff
    parameters when using paired data misfits and regularizations for a joint inversion.
    """

    chifact_target = 1.0
    beta_tol = 1e-1
    update_beta = True
    cooling_rate = 1
    cooling_factor = 2
    dmis_met = False

    @property
    def target(self):
        if getattr(self, "_target", None) is None:
            nD = np.array([survey.nD for survey in self.survey])

            self._target = nD * 0.5 * self.chifact_target

        return self._target

    @target.setter
    def target(self, val):
        self._target = val

    def initialize(self):
        self.dmis_met = np.zeros_like(self.invProb.betas, dtype=bool)

    def endIter(self):
        # Check if target misfit has been reached, if so, set dmis_met to True
        for i, phi_d in enumerate(self.invProb.phi_d_list):
            self.dmis_met[i] = phi_d < self.target[i]

        # check separately if misfits are within the tolerance,
        # otherwise, scale beta individually
        for i, phi_d in enumerate(self.invProb.phi_d_list):
            if self.opt.iter > 0 and self.opt.iter % self.cooling_rate == 0:
                target = self.target[i]
                ratio = phi_d / target
                if self.update_beta and ratio <= (1.0 + self.beta_tol):
                    if ratio <= 1:
                        ratio = np.maximum(0.75, ratio)
                    else:
                        ratio = np.minimum(1.5, ratio)

                    self.invProb.betas[i] /= ratio
                elif ratio > 1.0:
                    self.invProb.betas[i] /= self.cooling_factor

        self.reg.multipliers[:-1] = self.invProb.betas


class MovingAndMultiTargetStopping(InversionDirective):
    r"""
    Directive for setting stopping criteria for a joint inversion.
    Ensures both that all target misfits are met and there is a small change in the
    model. Computes the percentage change of the current model from the previous model.

    ..math::
        \frac {\| \mathbf{m_i} - \mathbf{m_{i-1}} \|} {\| \mathbf{m_{i-1}} \|}
    """

    tol = 1e-5
    beta_tol = 1e-1
    chifact_target = 1.0

    @property
    def target(self):
        if getattr(self, "_target", None) is None:
            nD = []
            for survey in self.survey:
                nD += [survey.nD]
            nD = np.array(nD)

            self._target = nD * 0.5 * self.chifact_target

        return self._target

    @target.setter
    def target(self, val):
        self._target = val

    def endIter(self):
        for phi_d, target in zip(self.invProb.phi_d_list, self.target):
            if np.abs(1.0 - phi_d / target) >= self.beta_tol:
                return
        if (
            np.linalg.norm(self.opt.xc - self.opt.x_last)
            / np.linalg.norm(self.opt.x_last)
            > self.tol
        ):
            return

        print(
            "stopping criteria met: ",
            np.linalg.norm(self.opt.xc - self.opt.x_last)
            / np.linalg.norm(self.opt.x_last),
        )
        self.opt.stopNextIteration = True



"""
============================================================================================
Induced Polarization (:mod:`SimPEG.electromagnetics.static.induced_polarization`)
============================================================================================
.. currentmodule:: SimPEG.electromagnetics.static.induced_polarization


Simulations
===========
.. autosummary::
  :toctree: generated/

  Simulation2DCellCentered
  Simulation2DNodal
  Simulation3DCellCentered
  Simulation3DNodal

Receivers, Sources, and Surveys
===============================
The ``induced_polarization`` module makes use of receivers, sources, and surveys
defined in the ``SimPEG.electromagnetics.static.resistivity`` module.
"""
from .simulation import (
    Simulation3DCellCentered,
    Simulation3DNodal,
    Simulation2DCellCentered,
    Simulation2DNodal,
)
from .survey import from_dc_to_ip_survey
from .run import run_inversion
from ..resistivity.survey import Survey
from ..resistivity import receivers
from ..resistivity import sources
from ..resistivity import utils

Simulation2DCellCentred = Simulation2DCellCentered
Simulation3DCellCentred = Simulation2DCellCentered



from ...survey import BaseRx
from ...utils import (
    validate_float,
    validate_string,
    validate_integer,
    validate_ndarray_with_shape,
)


#########################################
# POINT RECEIVER CLASS FOR VRM
#########################################


class Point(BaseRx):
    """Point receiver for viscous remanent magnetization simulations

    Parameters
    ----------
    locations : (n, 3) numpy.ndarray
        Receiver locations
    times : numpy.ndarray
        Time channels
    field_type : {'h', 'b', 'dhdt', 'dbdt'}
        Fields being measured.
    orientation : {'z', 'y', 'z'}
        Receiver orientation.
    """

    def __init__(
        self, locations=None, times=None, field_type=None, orientation="z", **kwargs
    ):
        if kwargs.pop("fieldType", None):
            raise AttributeError(
                "'fieldType' is a deprecated property. Please use 'field_type' instead."
                "'fieldType' be removed in SimPEG 0.17.0."
            )
        if field_type is None:
            raise AttributeError(
                "VRM receiver class cannot be instantiated witout 'field_type"
            )
        else:
            self.field_type = field_type

        super(Point, self).__init__(locations=locations, **kwargs)

        if times is None:
            raise AttributeError(
                "VRM receiver class cannot be instantiated without 'times'"
            )
        else:
            self.times = times

        self.orientation = orientation

    @property
    def times(self):
        """Time channels for the receivers

        Returns
        -------
        numpy.ndarray
            Time channels for the receivers
        """
        return self._times

    @times.setter
    def times(self, value):
        self._times = validate_ndarray_with_shape(
            "times", value, shape=("*",), dtype=float
        )

    @property
    def orientation(self):
        """Orientation of the receiver.

        Returns
        -------
        {'x', 'y', 'z'}
            Orientation of the receiver.
        """
        return self._orientation

    @orientation.setter
    def orientation(self, var):
        self._orientation = validate_string(
            "orientation", var, string_list=("x", "y", "z")
        )

    @property
    def field_type(self):
        """Field type; i.e. 'h', 'b', 'dhdt' or 'dbdt'

        Returns
        -------
        str
            Field type; i.e. 'h', 'b', 'dhdt' or 'dbdt'
        """
        return self._field_type

    @field_type.setter
    def field_type(self, var):
        var = validate_string(
            "field_type", var, string_list=("h", "b", "dhdt", "dbdt")
        ).lower()
        self._field_type = var

    @property
    def n_times(self):
        """Number of time channels

        Returns
        -------
        int
            Number of time channels
        """
        if self.times is not None:
            return len(self.times)

    @property
    def n_locations(self):
        """Number of receivers

        Returns
        -------
        int
            Number of receivers
        """
        return self.locations.shape[0]

    @property
    def nD(self):
        """Number of data associated with the receiver

        Returns
        -------
        int
            Number of data associated with the receiver
        """
        if self.times is not None:
            return self.locations.shape[0] * len(self.times)


class SquareLoop(Point):
    """Square loop receiver

    Measurements with this type of receiver are the field, integrated over the
    area of the loop, then multiplied by the number of coils, then normalized
    by the dipole moment. As a result, the units for fields predicted with this
    type of receiver are the same as 'h', 'b', 'dhdt' and 'dbdt', respectively.

    Parameters
    ----------
    locations : (n, 3) numpy.ndarray
        Center location of the square loop
    times : numpy.ndarray
        Time channels
    field_type : {'h', 'b', 'dhdt', 'dbdt'}
        Fields being measured.
    orientation : {'z', 'x', 'y'}
        Receiver orientation.
    width : float, default = 1.0
        Loop width (m)
    n_turns : int, default = 1
        Number of turns for the receiver coil
    quadrature_order : int, default = 3
        Order of numerical quadrature for approximating the magnetic flux through
        the receiver coil.
    """

    def __init__(
        self,
        locations=None,
        times=None,
        field_type=None,
        orientation="z",
        width=1.0,
        n_turns=1,
        quadrature_order=3,
        **kwargs,
    ):
        if "nTurns" in kwargs:
            raise AttributeError(
                "'nTurns' is a deprecated property. Please use 'n_turns' instead."
                "'nTurns' be removed in SimPEG 0.17.0."
            )

        if "quadOrder" in kwargs:
            raise AttributeError(
                "'quadOrder' is a deprecated property. Please use 'quadrature_order' instead."
                "'quadOrder' be removed in SimPEG 0.17.0."
            )

        super(SquareLoop, self).__init__(
            locations=locations,
            times=times,
            field_type=field_type,
            orientation=orientation,
            **kwargs,
        )

        self.width = width
        self.n_turns = n_turns
        self.quadrature_order = quadrature_order

    @property
    def width(self):
        """Square loop width

        Returns
        -------
        float
            Square loop width
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validate_float("width", value, min_val=0.0, inclusive_min=False)

    @property
    def n_turns(self):
        """Number of turns in the receiver coil

        Returns
        -------
        int
            Number of turns in the receiver coil
        """
        return self._n_turns

    @n_turns.setter
    def n_turns(self, value):
        self._n_turns = validate_integer("n_turns", value, min_val=1)

    @property
    def quadrature_order(self):
        """Quadrature order for determining flux through receiver coil

        Returns
        -------
        int
            Quadrature order for determining flux through receiver coil
        """
        return self._quadrature_order

    @quadrature_order.setter
    def quadrature_order(self, value):
        value = validate_integer("quadrature_order", value, min_val=1, max_val=7)
        self._quadrature_order = value



"""
====================================================================================================
Static Utilities (:mod:`SimPEG.electromagnetics.utils`)
====================================================================================================
.. currentmodule:: SimPEG.electromagnetics.static.utils


Electrode Utilities
===================
.. autosummary::
  :toctree: generated/

  electrode_separations
  pseudo_locations
  geometric_factor
  apparent_resistivity_from_voltage
  convert_survey_3d_to_2d_lines

Plotting Utilities
==================
.. autosummary::
  :toctree: generated/

  plot_pseudosection
  plot_3d_pseudosection


Survey Creation Utilities
=========================
.. autosummary::
  :toctree: generated/

  generate_survey_from_abmn_locations
  generate_dcip_survey
  generate_dcip_sources_line
  gen_3d_survey_from_2d_lines

Mesh Utilities
==============
.. autosummary::
  :toctree: generated/

  gettopoCC
  drapeTopotoLoc
  genTopography
  closestPointsGrid

"""
from .static_utils import (
    electrode_separations,
    source_receiver_midpoints,
    pseudo_locations,
    geometric_factor,
    apparent_resistivity_from_voltage,
    apparent_resistivity,
    plot_pseudosection,
    generate_dcip_survey,
    generate_dcip_survey_line,
    gen_DCIPsurvey,
    generate_dcip_sources_line,
    generate_survey_from_abmn_locations,
    writeUBC_DCobs,
    writeUBC_DClocs,
    convert_survey_3d_to_2d_lines,
    convertObs_DC3D_to_2D,
    readUBC_DC2Dpre,
    readUBC_DC3Dobs,
    xy_2_lineID,
    r_unit,
    getSrc_locs,
    gettopoCC,
    drapeTopotoLoc,
    genTopography,
    closestPointsGrid,
    gen_3d_survey_from_2d_lines,
    plot_1d_layer_model,
    plot_layer,
    plot_pseudoSection,
)

# Import if user has plotly
try:
    from .static_utils import plot_3d_pseudosection
except ImportError:
    pass
